# status
FAILED = "Failed"
UNNUMBERED = "Unnumbered"
SUCCESS = "Success"
LOADED = "Loaded"
NOT_LOADED = "Not Loaded"
# numbering schemes
KABAT = "kabat"
CHOTHIA = "chothia"
CHOTHIA_EXT = "chothia_ext"
MARTIN = "chothia_ext"
# servers
ABYSIS = "abysis"



from abpytools.core.formats.utils import get_protobuf_numbering_scheme, get_numbering_scheme_from_protobuf
from abpytools.core.chain import Chain


def _name_wrapper(func):
    """
    Internal usage only counter decorator to format sequence names.
    Args:
        func:

    Returns:

    """
    def helper(*args):
        helper.idi += 1
        name = func(*args)
        if isinstance(name, str):
            name = name.replace("%%IDI%%", str(helper.idi))
        else:
            names = list()
            for name_i in name:
                names.append(name_i.replace("%%IDI%%", str(helper.idi)))
            name = names
        return name
    helper.idi = 0
    return helper


@_name_wrapper
def _chain_name_getter(antibody_object):
    """
    Internal function that sets the name of Chain objects.
    This is required because it is possible to work with a ChainCollection
    where a Chain does not have a name. However, a name is
    required internally to store data on disk.

    Args:
        antibody_object (Chain):

    Returns:

    """
    # if antibody does not have name, generate name:
    # ID_chain_idi, where chain is heavy/light, idi is i = [1,..,N]
    if len(antibody_object.name) > 0:
        name = antibody_object.name
    else:
        name = f"ID_{antibody_object.chain}_%%IDI%%"
    return name


@_name_wrapper
def _fab_name_getter(name_i, light_chain, heavy_chain):
    """
    Internal function that sets the name of Fab objects
    when they are stored on disk.
    Args:
        name_i (str):
        light_chain (Chain):
        heavy_chain (Chain):

    Returns:

    """
    # if antibody does not have name, generate name:
    # ID_chain_idi, where chain is heavy/light, idi is i = [1,..,N]
    if len(name_i) > 0:
        names = name_i
    else:
        names = []
        for chain in [light_chain, heavy_chain]:
            if len(chain.name) > 0:
                names.append(chain.name)
            else:
                names.append(f"ID_{chain.chain}_%%IDI%%")
    return names


##############################################################################
#                                   SAVERS
##############################################################################

def json_FabCollection_formatter(fab_object):
    """
    Internal function to serialise FabCollection objects in JSON format
    Args:
        fab_object (FabCollection):

    Returns:

    """

    light_antibody_json = json_ChainCollection_formatter(fab_object._light_chains)
    heavy_antibody_json = json_ChainCollection_formatter(fab_object._heavy_chains)

    fab_data = dict()
    ordered_names = []

    for light_antibody, heavy_antibody in zip(
            light_antibody_json['ordered_names'], heavy_antibody_json['ordered_names']):
        name = f"{light_antibody_json[light_antibody]['name']}-" \
               f"{heavy_antibody_json[heavy_antibody]['name']}"
        fab_data[name] = [light_antibody_json[light_antibody],
                          heavy_antibody_json[heavy_antibody]]
        ordered_names.append(name)
    fab_data['ordered_names'] = ordered_names

    return fab_data


def json_ChainCollection_formatter(chain_objects):
    """
    Internal function to serialise ChainCollection objects in JSON format
    Args:
        chain_objects (ChainCollection):

    Returns:

    """
    data = dict()
    ordered_names = []
    # reset wrapper counter
    _chain_name_getter.idi = 0
    for antibody in chain_objects:
        name = _chain_name_getter(antibody)
        ordered_names.append(name)
        antibody_dict = antibody.ab_format()
        # use name generated from _chain_name_getter instead
        antibody_dict['name'] = name
        data[name] = antibody_dict
    data['ordered_names'] = ordered_names
    return data


def pb2_ChainCollection_formatter(chain_objects, proto_parser, reset_status=True):
    """
    Internal function to serialise a ChainCollection object to .pb2 format
    according to definitition in 'format/chain.proto'.

    Args:
        chain_objects (ChainCollection):
        proto_parser (ChainCollectionProto):
        reset_status (bool):

    Returns:

    """
    # reset wrapper counter
    if reset_status:
        _chain_name_getter.idi = 0
    for chain_object in chain_objects:
        pb2_add_chain(chain_object, proto_parser)


def pb2_FabCollection_formatter(fab_object, proto_parser, reset_status=True):
    """
    Internal function to serialise a FabCollection object to .pb2 format
    according to definitition in 'format/fab.proto'.

    Args:
        fab_object (FabCollection):
        proto_parser (FabCollectionProto):
        reset_status (bool):

    Returns:

    """
    # reset wrapper counter
    if reset_status:
        _chain_name_getter.idi = 0
    for name_i, light_chain_i, heavy_chain_i in zip(fab_object.names,
                                                    fab_object._light_chains,
                                                    fab_object._heavy_chains):

        proto_fab = proto_parser.fabs.add()
        name_i = _fab_name_getter(name_i, light_chain_i, heavy_chain_i)
        proto_fab.name = name_i
        add_Chain_to_protobuf(light_chain_i, proto_fab.light_chain)
        add_Chain_to_protobuf(heavy_chain_i, proto_fab.heavy_chain)


def pb2_add_chain(chain_object, proto_parser):
    """
    Populates a protobuf ProtoChain message from Chain object
    and adds it to ChainCollectionProto
    Args:
        chain_object (Chain):
        proto_parser (ChainCollectionProto):

    Returns:

    """
    proto_antibody = proto_parser.chains.add()
    add_Chain_to_protobuf(chain_object, proto_antibody)


def add_Chain_to_protobuf(antibody_obj, proto_obj):
    """
    Helper function to populate a ProtoChain message.
    Args:
        antibody_obj (Chain):
        proto_obj (ChainProto):

    Returns:

    """
    proto_obj.name = _chain_name_getter(antibody_obj)
    proto_obj.sequence = antibody_obj.sequence
    proto_obj.numbering_scheme = get_protobuf_numbering_scheme(antibody_obj.numbering_scheme)
    proto_obj.numbering.extend(antibody_obj.numbering)
    proto_obj.chain_type = antibody_obj.chain
    proto_obj.MW = antibody_obj.mw
    for x in {**antibody_obj.ab_regions()[0], **antibody_obj.ab_regions()[1]}.items():
        proto_obj_cdr = proto_obj.region.add()
        proto_obj_cdr.region_name = x[0]
        proto_obj_cdr.region_positions.extend(x[1])
    proto_obj.pI = antibody_obj.pI


##############################################################################
#                                   PARSERS
##############################################################################
def fasta_ChainCollection_parser(raw_fasta, numbering_scheme):

    names = list()
    sequences = list()
    antibody_objects = list()
    for line in raw_fasta:
        if line.startswith(">"):
            names.append(line.replace("\n", "")[1:])
        # if line is empty skip line
        elif line.isspace():
            pass
        else:
            sequences.append(line.replace("\n", ""))
    if len(names) != len(sequences):
        raise ValueError("Error reading file: make sure it is FASTA format")

    for name, sequence in zip(names, sequences):
        antibody_objects.append(Chain(name=name, sequence=sequence,
                                      numbering_scheme=numbering_scheme))

    return antibody_objects


def pb2_FabCollection_parser(proto_parser):
    from abpytools.core.fab import Fab
    fab_objects = list()
    for fab_i in proto_parser.fabs:

        light_chain = pb2_Chain_parser(fab_i.light_chain)
        heavy_chain = pb2_Chain_parser(fab_i.heavy_chain)

        fab = Fab(light_chain=light_chain, heavy_chain=heavy_chain, name=fab_i.name)

        fab_objects.append(fab)

    return fab_objects


def pb2_ChainCollection_parser(proto_parser):
    antibody_objects = list()
    names = list()
    for chain_i in proto_parser.chains:

        antibody_i = pb2_Chain_parser(chain_i)

        antibody_i._loading_status = 'Loaded'
        names.append(chain_i.name)
        antibody_objects.append(antibody_i)

    return antibody_objects


def pb2_Chain_parser(proto_chain):
    """
    Populate Chain object from protobuf file

    Args:
        proto_chain (ChainProto):

    Returns:

    """
    chain_obj = Chain(name=proto_chain.name, sequence=proto_chain.sequence)

    if proto_chain.numbering_scheme:
        chain_obj._numbering_scheme = get_numbering_scheme_from_protobuf(proto_chain.numbering_scheme)

    if proto_chain.numbering:
        # convert google.protobuf.pyext._message.RepeatedScalarContainer to list
        chain_obj.numbering = list(proto_chain.numbering)
    else:
        chain_obj.numbering = chain_obj.ab_numbering()

    if proto_chain.chain_type:
        chain_obj._chain = proto_chain.chain_type
    else:
        chain_obj._chain = Chain.determine_chain_type(proto_chain.numbering)

    if proto_chain.MW:
        chain_obj.mw = proto_chain.MW
    else:
        chain_obj.mw = chain_obj.ab_molecular_weight()

    if proto_chain.region:
        regions = dict()
        for region in proto_chain.region:
            regions[region.region_name] = region.region_positions
        chain_obj.CDR = regions
    else:
        chain_obj.CDR = chain_obj.ab_regions()

    if proto_chain.pI:
        chain_obj.pI = proto_chain.pI
    else:
        chain_obj.pI = chain_obj.ab_pi()

    return chain_obj


def json_FabCollection_parser(raw_data):
    from abpytools.core.fab import Fab
    fab_objects = list()

    ordered_names = raw_data.pop('ordered_names')

    for key_i in ordered_names:
        light_chain_i = json_Chain_parser(raw_data[key_i][0], raw_data[key_i][0]["name"])
        heavy_chain_i = json_Chain_parser(raw_data[key_i][1], raw_data[key_i][1]["name"])
        fab_i = Fab(light_chain=light_chain_i, heavy_chain=heavy_chain_i, name=key_i)
        fab_objects.append(fab_i)

    return fab_objects


def json_ChainCollection_parser(raw_data):

    antibody_objects = list()

    ordered_names = raw_data.pop('ordered_names')

    for key_i in ordered_names:
        chain_i = json_Chain_parser(raw_data[key_i], key_i)
        antibody_objects.append(chain_i)

    return antibody_objects


def json_Chain_parser(antibody_dict, name):
    antibody_i = Chain(name=name, sequence=antibody_dict['sequence'])
    antibody_i.numbering = antibody_dict['numbering']
    antibody_i._chain = antibody_dict['chain']
    antibody_i.mw = antibody_dict['MW']
    antibody_i.CDR = antibody_dict["CDR"]
    antibody_i._numbering_scheme = antibody_dict["numbering_scheme"]
    antibody_i.pI = antibody_dict["pI"]
    antibody_i._loading_status = 'Loaded'
    return antibody_i



from ..core.chain_collection import ChainCollection
import seaborn as sns
from .analysis_helper_functions import switch_interactive_mode
from matplotlib import pyplot as plt
from scipy.cluster.hierarchy import linkage, dendrogram
import scipy.spatial.distance as ssd
from ..utils.python_config import PythonConfig


class DistancePlot(ChainCollection):
    def __init__(self, antibody_objects=None, path=None):
        super().__init__(antibody_objects=antibody_objects, path=path)

    def plot_heatmap(self, feature='chou', distance_metric='cosine_distance', save=False, ax=None, labels=None,
                     multiprocessing=False, file_name='./heatmap.png', **kwargs):

        data = self.distance_matrix(feature=feature, metric=distance_metric, multiprocessing=multiprocessing)

        switch_interactive_mode(save=save)

        if ax is None:
            f, ax = plt.subplots(1, 1, figsize=(8, 6))
            ax.set(xlabel='Antibody', ylabel='Antibody', title=distance_metric,
                   xticks=range(self.n_ab), yticks=range(self.n_ab))

        if labels is None:
            labels = self.names

        sns.heatmap(data, ax=ax, **kwargs)
        ax.set_yticklabels(labels, rotation='horizontal')
        ax.set_xticklabels(labels, rotation=60)

        ipython_config = PythonConfig()
        if ipython_config.ipython_info == 'notebook' and save is False:
            plt.plot()
        else:
            plt.savefig(file_name)

    def plot_dendrogram(self, feature='chou', distance_metric='cosine_distance', save=False, ax=None, labels=None,
                        multiprocessing=False, **kwargs):

        switch_interactive_mode(save=save)

        data = self.distance_matrix(feature=feature, metric=distance_metric, multiprocessing=multiprocessing)
        # convert the redundant n*n square matrix form into a condensed nC2 array
        data = ssd.squareform(data)

        clustered_data = linkage(y=data)

        if ax is None:
            f, ax = plt.subplots(1, 1, figsize=(8, 6))
            ax.set(xlabel='Antibody', ylabel='Distance', title=distance_metric)

        if labels is None:
            labels = self.names

        # plot dendrogram
        _ = dendrogram(clustered_data, labels=labels, ax=ax, **kwargs)

        ipython_config = PythonConfig()
        if ipython_config.ipython_info == 'notebook' and save is False:
            plt.plot()
        else:
            plt.savefig(file_name)



import cython

cdef extern from "<malloc.h>" nogil:
    void *memalign(size_t alignment, size_t size);
    int posix_memalign(void **, size_t, size_t);

ctypedef fused scalar_or_char:
    cython.floating
    cython.integral
    cython.char

ctypedef fused scalar:
    cython.floating
    cython.integral

cdef double* get_C_double_array_pointers(list a, int size)
cdef double** get_C_double_array_pp(double* a, int size)
cdef char* get_C_char_array_pointers(str a, int size)
cdef double** get_pp_from_ptr(double* ptr, int size)
cdef void get_array_from_ptr(double* ptr, double* a_, int size)
cdef void release_C_pointer(scalar_or_char *a)
cdef void release_C_pp(scalar_or_char** a)
cdef double** allocate_pp(int size)


cdef class Matrix2D_backend:
    cdef public int n_cols, n_rows, size_
    cdef list values_
    cdef double* data_C

    cdef double* _get_row_pointer(self, int row)
    cdef double* _get_value_pointer(self, int row, int column)
    cpdef list _get_row(self, int row)
    cpdef list _get_values(self)
    cdef double _get_value(self, int row, int col)
    cdef void _set_value(self, int row, int col, double value)

cdef class Vector:
    cdef int size_, derived_
    cdef double* data_C

    # additional instantiation methods
    @staticmethod
    cdef Vector create(double* ptr, int size)
    @staticmethod
    cdef Vector allocate(int size)
    @staticmethod
    cdef Vector allocate_with_value(int size, double value)

    cdef double _get_element(self, int idx)
    cdef void _set_array_value(self, int idx, double value)
    cpdef double dot_product(self, Vector other)
    cpdef Vector add(self, Vector other)
    cpdef Vector subtract(self, Vector other)
    cpdef Vector multiply(self, Vector other)
    cpdef Vector exp(self)
    cpdef double norm(self, int p=*)


cdef double internal_vector_dot_product_(double *u, double *v, int size)
cdef double internal_vector_norm_(double *u, int norm, int size)



# from abpytools.analysis.cdr_length import CDRLength
# from abpytools.analysis.amino_acid_freq import AminoAcidFreq
# from abpytools.analysis.sequence_alignment import SequenceAlignment
# from abpytools.analysis.cluster import Cluster



from .chain import Chain
import numpy as np
import logging
from abpytools.utils import PythonConfig, Download
import json
import os
import pandas as pd
from .helper_functions import numbering_table_sequences, numbering_table_region, numbering_table_multiindex
from operator import itemgetter
from urllib import parse
from math import ceil
from .base import CollectionBase
from ..features.composition import *
from ..analysis.distance_metrics import *
from ..core.cache import Cache
from multiprocessing import Manager, Process
from inspect import signature
from .utils import (json_ChainCollection_formatter, pb2_ChainCollection_formatter, pb2_ChainCollection_parser,
                    fasta_ChainCollection_parser, json_ChainCollection_parser)
from .flags import *

# setting up debugging messages
logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.DEBUG)

ipython_config = PythonConfig()
if ipython_config.ipython_info == 'notebook':
    from tqdm import tqdm_notebook as tqdm  # pragma: no cover
else:
    from tqdm import tqdm

if BACKEND_FLAGS.HAS_PROTO:
    from abpytools.core.formats import ChainCollectionProto


class ChainCollection(CollectionBase):
    """
    Object containing Chain objects and to perform analysis on the ensemble.

    """

    def __init__(self, antibody_objects=None, load=True, **kwargs):
        """

        Args:
            antibody_objects:
            load:
            **kwargs:
        """

        if antibody_objects is None:
            self.antibody_objects = []
        else:
            if isinstance(antibody_objects, ChainCollection):
                antibody_objects = antibody_objects.antibody_objects
            elif not isinstance(antibody_objects, list):
                raise ValueError("Expected a list, instead got object of type {}".format(type(antibody_objects)))

            elif not all(isinstance(obj, Chain) for obj in antibody_objects) and len(antibody_objects) > 0:
                raise ValueError("Expected a list containing objects of type Chain")

            self.antibody_objects = antibody_objects

            if len(set(x.numbering_scheme for x in antibody_objects)) == 1:
                self._numbering_scheme = antibody_objects[0].numbering_scheme
            else:
                raise ValueError("ChainCollection only support Chain objects with the same numbering scheme.")

            if len(set(x.chain for x in antibody_objects)) == 1:
                self._chain = antibody_objects[0].chain
            elif len(set(x.chain for x in antibody_objects)) == 0:
                self._chain = ''
            else:
                raise ValueError("ChainCollection only support Chain objects with the same chain type.")

            if load:
                self.load(**kwargs)

    def load(self, show_progressbar=True, n_threads=4, verbose=True):
        self.antibody_objects, self._chain = load_from_antibody_object(
            antibody_objects=self.antibody_objects,
            show_progressbar=show_progressbar,
            n_threads=n_threads, verbose=verbose)

    @classmethod
    def load_from_fasta(cls, path, numbering_scheme=NUMBERING_FLAGS.CHOTHIA, n_threads=20,
                        verbose=True, show_progressbar=True):
        if not os.path.isfile(path):
            raise ValueError("File does not exist!")
        with open(path, 'r') as f:
            antibody_objects = fasta_ChainCollection_parser(f, numbering_scheme=numbering_scheme)

        chain_collection = cls(antibody_objects=antibody_objects, load=True,
                               n_threads=n_threads, verbose=verbose,
                               show_progressbar=show_progressbar)

        return chain_collection

    @classmethod
    def load_from_pb2(cls, path, n_threads=20, verbose=True, show_progressbar=True):
        with open(path, 'rb') as f:
            proto_parser = ChainCollectionProto()
            proto_parser.ParseFromString(f.read())

        antibody_objects = pb2_ChainCollection_parser(proto_parser)

        chain_collection = cls(antibody_objects=antibody_objects, load=True,
                               n_threads=n_threads, verbose=verbose,
                               show_progressbar=show_progressbar)

        return chain_collection

    @classmethod
    def load_from_json(cls, path, n_threads=20, verbose=True, show_progressbar=True):

        with open(path, 'r') as f:
            data = json.load(f)

        antibody_objects = json_ChainCollection_parser(data)

        chain_collection = cls(antibody_objects=antibody_objects, load=True,
                               n_threads=n_threads, verbose=verbose,
                               show_progressbar=show_progressbar)

        return chain_collection

    def save_to_json(self, path, update=True):
        with open(os.path.join(path + '.json'), 'w') as f:
            data = json_ChainCollection_formatter(self.antibody_objects)
            json.dump(data, f, indent=2)

    def save_to_pb2(self, path, update=True):
        proto_parser = ChainCollectionProto()
        try:
            with open(os.path.join(path + '.pb2'), 'rb') as f:
                proto_parser.ParseFromString(f.read())
        except IOError:
            # print("Creating new file")
            pass

        pb2_ChainCollection_formatter(self.antibody_objects, proto_parser)

        with open(os.path.join(path + '.pb2'), 'wb') as f:
            f.write(proto_parser.SerializeToString())

    def save_to_fasta(self, path, update=True):
        with open(os.path.join(path + '.fasta'), 'w') as f:
            f.writelines(make_fasta(self.names, self.sequences))

    def molecular_weights(self, monoisotopic=False):

        """

        :param monoisotopic: bool whether to use monoisotopic values
        :return: list
        """

        return [x.ab_molecular_weight(monoisotopic=monoisotopic) for x in self.antibody_objects]

    def extinction_coefficients(self, extinction_coefficient_database='Standard', reduced=False):

        """

        :param extinction_coefficient_database: string with the name of the database to use
        :param reduced: bool whether to consider the cysteines to be reduced
        :return: list
        """

        return [x.ab_ec(extinction_coefficient_database=extinction_coefficient_database,
                        reduced=reduced) for x in self.antibody_objects]

    def hydrophobicity_matrix(self):

        if self._chain == CHAIN_FLAGS.HEAVY_CHAIN:
            num_columns = 158
        else:
            num_columns = 138
        abs_hydrophobicity_matrix = np.zeros((len(self.antibody_objects), num_columns))

        for row in range(abs_hydrophobicity_matrix.shape[0]):
            abs_hydrophobicity_matrix[row] = self.antibody_objects[row].hydrophobicity_matrix

        return abs_hydrophobicity_matrix

    def get_object(self, name=''):

        """

        :param name: str
        :return:
        """

        if name in self.names:
            index = self.names.index(name)
            return self[index]
        else:
            raise ValueError('Could not find sequence with specified name')

    def ab_region_index(self):

        """
        method to determine index of amino acids in CDR regions
        :return: dictionary with names as keys and each value is a dictionary with keys CDR and FR
        'CDR' entry contains dictionaries with CDR1, CDR2 and CDR3 regions
        'FR' entry contains dictionaries with FR1, FR2, FR3 and FR4 regions
        """
        return {x.name: {'CDR': x.ab_regions()[0], 'FR': x.ab_regions()[1]} for x in self.antibody_objects}

    def numbering_table(self, as_array=False, region='all'):

        region = numbering_table_region(region)

        table = np.row_stack(
            [x.ab_numbering_table(as_array=True, region=region) for x in self.antibody_objects])

        if as_array:
            return table

        else:
            # return the data as a pandas.DataFrame -> it's slower but looks nicer and makes it easier to get
            # the data of interest
            whole_sequence_dict, whole_sequence = numbering_table_sequences(region, self._numbering_scheme, self._chain)

            multi_index = numbering_table_multiindex(region=region,
                                                     whole_sequence_dict=whole_sequence_dict)

            # create the DataFrame and assign the columns and index names
            data = pd.DataFrame(data=table)
            data.columns = multi_index
            data.index = self.names
            return data

    def igblast_server_query(self, chunk_size=50, show_progressbar=True, **kwargs):
        """

        :param show_progressbar:
        :param chunk_size:
        :param kwargs: keyword arguments to pass to igblast_options
        :return:
        """

        # check if query is larger than 50 sequences
        # if so split into several queries
        query_list = self._split_to_chunks(chunk_size=chunk_size)
        n_chunks = ceil(len(self) / chunk_size) - 1

        if show_progressbar:
            for query in tqdm(query_list, total=n_chunks):
                self._igblast_server_query(query, **kwargs)

        else:
            for query in query_list:
                self._igblast_server_query(query, **kwargs)

    def _igblast_server_query(self, query, **kwargs):

        # prepare raw data
        fasta_query = make_fasta(names=query.names, sequences=query.sequences)

        # get url with igblast options
        url = igblast_options(sequences=fasta_query, **kwargs)

        # send and download query
        q = Download(url, verbose=False)

        try:
            q.download()
        except ValueError:  # pragma: no cover
            raise ValueError("Check the internet connection.")  # pragma: no cover

        igblast_result = q.html

        self._parse_igblast_query(igblast_result, query.names)

    def igblast_local_query(self, file_path):

        # load in file
        with open(file_path, 'r') as f:
            igblast_result = f.readlines()

        self._parse_igblast_query(igblast_result, self.names)

    def append(self, antibody_obj):

        self.antibody_objects += antibody_obj

    def pop(self, index=-1):

        if index > len(self):
            raise ValueError("The given index is outside the range of the object.")

        element_to_pop = self[index]

        self._destroy(index=index)

        return element_to_pop

    def _destroy(self, index):

        del self.antibody_objects[index]

    # def filter(self):
    #
    #     # TODO: complete method
    #     pass
    #
    def set_numbering_scheme(self, numbering_scheme, realign=True):
        if realign:
            try:
                self._numbering_scheme = numbering_scheme
                self.antibody_objects, self._chain = load_from_antibody_object(self.antibody_objects)
            except:
                print("Could not realign sequences, nothing has been changed.")
        else:
            self._numbering_scheme = numbering_scheme

    @property
    def names(self):
        return [x.name for x in self.antibody_objects]

    @property
    def sequences(self):
        return [x.sequence for x in self.antibody_objects]

    @property
    def aligned_sequences(self):
        return [x.aligned_sequence for x in self.antibody_objects]

    @property
    def n_ab(self):
        return len(self.sequences)

    @property
    def chain(self):
        if self._chain == '':
            chains = set([x.chain for x in self.antibody_objects])
            if len(chains) == 1:
                self._chain = next(iter(chains))
                return self._chain
            else:
                raise ValueError('Different types of chains found in collection!')
        else:
            return self._chain

    @property
    def numbering_scheme(self):
        return self._numbering_scheme

    @property
    def charge(self):
        return np.array([x.ab_charge() for x in self.antibody_objects])

    @property
    def total_charge(self):
        return {x.name: x.ab_total_charge() for x in self.antibody_objects}

    @property
    def germline_identity(self):
        return {x.name: x.germline_identity for x in self.antibody_objects}

    @property
    def germline(self):
        return {x.name: x.germline for x in self.antibody_objects}

    def _string_summary_basic(self):
        return "abpytools.ChainCollection Chain type: {}, Number of sequences: {}".format(self._chain,
                                                                                          len(self.antibody_objects))

    def __repr__(self):
        return "<%s at 0x%02x>" % (self._string_summary_basic(), id(self))

    def __len__(self):
        return len(self.antibody_objects)

    def __getitem__(self, indices):
        if isinstance(indices, int):
            return self.antibody_objects[indices]
        else:
            return ChainCollection(antibody_objects=list(itemgetter(*indices)(self.antibody_objects)))

    def __add__(self, other):
        if isinstance(other, ChainCollection):
            if self.numbering_scheme != other.numbering_scheme:
                raise ValueError("Concatenation requires ChainCollection "
                                 "objects to use the same numbering scheme.")
            else:
                new_object_list = self.antibody_objects + other.antibody_objects

        elif isinstance(other, Chain):
            if self.numbering_scheme != other.numbering_scheme:
                raise ValueError("Concatenation requires Chain object to use "
                                 "the same numbering scheme as ChainCollection.")
            else:
                new_object_list = self.antibody_objects + [other]

        else:
            raise ValueError("Concatenation requires other to be of type "
                             "ChainCollection, got {} instead".format(type(other)))
        return ChainCollection(antibody_objects=new_object_list, load=False)

    def _split_to_chunks(self, chunk_size=50):
        """
        Helper function to split ChainCollection into size chunk_size and returns generator
        :param chunk_size: int, size of each chunk
        :return: generator to iterate of each chunk of size chunk_size
        """

        if self.n_ab > chunk_size:

            for x in range(0, self.n_ab, chunk_size):
                yield self[range(x, min(x + chunk_size, self.n_ab))]

        else:
            yield self

    def _parse_igblast_query(self, igblast_result, names):

        igblast_result_dict = load_igblast_query(igblast_result, names)

        # unpack results
        for name in names:
            obj_i = self.get_object(name=name)
            obj_i.germline = igblast_result_dict[name][1]
            obj_i.germline_identity = igblast_result_dict[name][0]

    def loading_status(self):
        return [x.status for x in self.antibody_objects]

    def composition(self, method='count'):
        """
        Amino acid composition of each sequence. Each resulting list is organised alphabetically (see composition.py)
        :param method:
        :return:
        """
        if method == 'count':
            return [order_seq(aa_composition(seq)) for seq in self.sequences]
        elif method == 'freq':
            return [order_seq(aa_frequency(seq)) for seq in self.sequences]
        elif method == 'chou':
            return chou_pseudo_aa_composition(self.sequences)
        elif method == 'triad':
            return triad_method(self.sequences)
        elif method == 'hydrophobicity':
            return self.hydrophobicity_matrix()
        elif method == 'volume':
            return side_chain_volume(self.sequences)
        else:
            raise ValueError("Unknown method")

    def distance_matrix(self, feature=None, metric='cosine_similarity', multiprocessing=False):

        """
        Returns the distance matrix using a given feature and distance metric
        :param feature: string with the name of the feature to use
        :param metric: string with the name of the metric to use
        :param multiprocessing: bool to turn multiprocessing on/off (True/False)
        :return: list of lists with distances between all sequences of len(data) with each list of len(data)
                 when i==j M_i,j = 0
        """

        if feature is None:
            transformed_data = self.sequences
        elif isinstance(feature, str):
            # in this case the features are calculated using a predefined featurisation method (see self.composition)
            transformed_data = self.composition(method=feature)

        elif isinstance(feature, list):
            # a user defined list with vectors
            if len(feature) != self.n_ab:
                raise ValueError("Expected a list of size {}, instead got {}.".format(self.n_ab, len(feature)))
            else:
                transformed_data = feature
        else:
            raise TypeError("Unexpected input for feature argument.")

        if metric == 'cosine_similarity':
            distances = self._run_distance_matrix(transformed_data, cosine_similarity, multiprocessing=multiprocessing)

        elif metric == 'cosine_distance':
            distances = self._run_distance_matrix(transformed_data, cosine_distance, multiprocessing=multiprocessing)

        elif metric == 'hamming_distance':
            # be careful hamming distance only works when all sequences have the same length
            distances = self._run_distance_matrix(transformed_data, hamming_distance, multiprocessing=multiprocessing)

        elif metric == 'levenshtein_distance':
            distances = self._run_distance_matrix(transformed_data, levenshtein_distance,
                                                  multiprocessing=multiprocessing)

        elif metric == 'euclidean_distance':
            distances = self._run_distance_matrix(transformed_data, euclidean_distance, multiprocessing=multiprocessing)

        elif metric == 'manhattan_distance':
            distances = self._run_distance_matrix(transformed_data, manhattan_distance, multiprocessing=multiprocessing)

        elif callable(metric):
            # user defined metric function
            user_function_signature = signature(metric)

            # number of params should be two, can take args with defaults though
            default_params = sum(['=' in x for x in user_function_signature.parameters])

            if len(user_function_signature.parameters) - default_params > 2:
                raise ValueError("Expected a function with two parameters")
            else:
                distances = self._run_distance_matrix(transformed_data, metric, multiprocessing=multiprocessing)

        else:
            raise ValueError("Unknown distance metric.")

        return distances

    def _run_distance_matrix(self, data, metric, multiprocessing=False):

        """
        Helper function to setup the calculation of each entry in the distance matrix
        :param data: list with all sequences
        :param metric: function that takes two string and calculates distance
        :param multiprocessing: bool to turn multiprocessing on/off (True/False)
        :return: list of lists with distances between all sequences of len(data) with each list of len(data)
                 when i==j M_i,j = 0
        """

        if multiprocessing:
            with Manager() as manager:
                cache = manager.dict()
                matrix = manager.dict()

                jobs = [Process(target=self._distance_matrix,
                                args=(data, i, metric, cache, matrix)) for i in range(len(data))]

                for j in jobs:
                    j.start()
                for j in jobs:
                    j.join()

                # order the data
                return [matrix[x] for x in range(len(data))]

        else:
            cache = Cache(max_cache_size=(len(data) * (len(data) - 1)) / 2)
            matrix = Cache(max_cache_size=len(data))

            for i in range(len(data)):
                cache.update(i, self._distance_matrix(data, i, metric, cache, matrix))

            return [matrix[x] for x in range(len(data))]

    @staticmethod
    def _distance_matrix(data, i, metric, cache, matrix):

        """
        Function to calculate distance from the ith sequence of the ith row to the remaining entries in the same row
        :param data: list with all sequences
        :param i: int that indicates the matrix row being processed
        :param metric: function that takes two string and calculates distance
        :param cache: either a Manager or Cache object to cache results
        :param matrix: either a Manager or Cache object to store results in a matrix
        :return: None
        """

        row = []
        seq_1 = data[i]
        for j, seq_2 in enumerate(data):

            if i == j:
                row.append(0)
                continue

            keys = ('{}-{}'.format(i, j), '{}-{}'.format(j, i))
            if keys[0] not in cache or keys[1] not in cache:
                cache['{}-{}'.format(i, j)] = metric(seq_1, seq_2)
            if keys[0] in cache:
                row.append(cache[keys[0]])
            elif keys[1] in cache:
                row.append(cache[keys[0]])
            else:
                raise ValueError("Bug in row {} and column {}".format(i, j))

        matrix[i] = row


def load_antibody_object(antibody_object):
    antibody_object.load()
    return antibody_object


def load_from_antibody_object(antibody_objects, show_progressbar=True, n_threads=20, verbose=True):
    """

    Args:
        antibody_objects (list):
        show_progressbar (bool):
        n_threads (int):
        verbose (bool):

    Returns:

    """

    if verbose:
        print("Loading in antibody objects")

    from queue import Queue
    import threading

    q = Queue()
    for i in range(n_threads):
        t = threading.Thread(target=worker, args=(q,))
        t.daemon = True
        t.start()

    if show_progressbar:
        for antibody_object in tqdm(antibody_objects):
            q.put(antibody_object)

    else:
        for antibody_object in antibody_objects:
            q.put(antibody_object)

    q.join()

    # if show_progressbar:
    #     aprun = parallelexecutor(use_bar='tqdm', n_jobs=n_jobs, timeout=timeout)
    # else:
    #     aprun = parallelexecutor(use_bar='None', n_jobs=n_jobs, timeout=timeout)
    #
    # # load in objects in parallel
    # antibody_objects = aprun(total=len(antibody_objects))(
    #     delayed(load_antibody_object)(obj) for obj in antibody_objects)

    status = [x.status for x in antibody_objects]
    failed = sum([1 if x == 'Not Loaded' or x == 'Failed' else 0 for x in status])

    # remove objects that did not load
    while 'Not Loaded' in status:
        i = status.index('Not Loaded')
        del antibody_objects[i], status[i]

    while 'Failed' in status:
        i = status.index('Failed')
        del antibody_objects[i], status[i]

    if verbose:
        print("Failed to load {} objects in list".format(failed))

    loaded_obj_chains = [x.chain for x in antibody_objects if x.status == 'Loaded']

    if len(set(loaded_obj_chains)) == 1:
        chain = loaded_obj_chains[0]
    else:
        raise ValueError("All sequences must be of the same chain type: Light or Heavy",
                         set([x.chain for x in loaded_obj_chains]))

    n_ab = len(loaded_obj_chains)

    if n_ab == 0:
        raise ValueError("Could not find any heavy or light chains in provided file or list of objects")

    return antibody_objects, chain


def load_igblast_query(igblast_result, names):
    """

    :param names:
    :param igblast_result:
    :return:
    """

    try:
        from bs4 import BeautifulSoup
    except ImportError:
        raise ImportError("Please install bs4 to parse the IGBLAST html file:"
                          "pip install beautifulsoup4")

    # instantiate BeautifulSoup object to make life easier with the html text!
    if isinstance(igblast_result, list):
        soup = BeautifulSoup(''.join(igblast_result), "lxml")
    else:
        soup = BeautifulSoup(igblast_result, "lxml")

    # get the results found in <div id="content"> and return the text as a string
    results = soup.find(attrs={'id': "content"}).text

    # get query names
    query = re.compile('Query: (.*)')
    query_ids = query.findall(results)

    # make sure that all the query names in query are in self.names
    if not set(names).issubset(set(query_ids)):
        raise ValueError('Make sure that you gave the same names in ChainCollection as you gave'
                         'in the query submitted to IGBLAST')

    # regular expression to get tabular data from each region
    all_queries = re.compile('(Query: .*?)\n\n\n\n', re.DOTALL)

    # parse the results with regex and get a list with each query data
    parsed_results = all_queries.findall(results)

    # regex to get the FR and CDR information for each string in parsed results
    region_finder = re.compile('^([CDR\d|FR\d|Total].*)', re.MULTILINE)

    result_dict = {}

    # iterate over each string in parsed result which contains the result for individual queries
    for query_result in parsed_results:

        # get query name and get the relevant object
        query_i = query.findall(query_result)[0]

        # check if the query being parsed is part of the object
        # (not all queries have to be part of the object, but the object names must be a subset of the queries)

        if query_i not in set(names):
            continue

        # list with CDR and FR info for query result
        region_info = region_finder.findall(query_result)

        # get the data from region info with dict comprehension
        germline_identity = {x.split()[0].split('-')[0]: float(x.split()[-1]) for x in region_info}

        # get the top germline assignment
        v_line_assignment = re.compile('V\s{}\t.*'.format(query_i))

        # the top germline assignment is at the top (index 0)
        germline_result = v_line_assignment.findall(results)[0].split()

        # store the germline assignment and the bit score in a tuple as the germline attribute of Chain
        germline = (germline_result[2], float(germline_result[-2]))

        result_dict[query_i] = (germline_identity, germline)

    return result_dict


def worker(q):
    while True:
        item = q.get()
        load_antibody_object(item)
        q.task_done()


def make_fasta(names, sequences):
    file_string = ''
    for name, sequence in zip(names, sequences):
        file_string += '>{}\n'.format(name)
        file_string += '{}\n'.format(sequence)

    return file_string


def igblast_options(sequences, domain='imgt',
                    germline_db_V='IG_DB/imgt.Homo_sapiens.V.f.orf.p',
                    germline_db_D='IG_DB/imgt.Homo_sapiens.D.f.orf',
                    germline_db_J='IG_DB / imgt.Homo_sapiens.J.f.orf',
                    num_alignments_V=1, num_alignments_D=1, num_alignments_J=1):
    values = {"queryseq": sequences,
              "germline_db_V": germline_db_V,
              "germline_db_D": germline_db_D,
              "germline_db_J": germline_db_J,
              "num_alignments_V": str(num_alignments_V),
              "num_alignments_D": str(num_alignments_D),
              "num_alignments_J": str(num_alignments_J),
              "outfmt": "7",
              "domain": domain,
              "program": "blastp"}

    url = "http://www.ncbi.nlm.nih.gov/igblast/igblast.cgi?"
    url += parse.urlencode(values)

    return url



import re
import numpy as np
from ..utils import DataLoader, Download, NumberingException
import pandas as pd
from .helper_functions import numbering_table_sequences, numbering_table_region, numbering_table_multiindex
from . import Cache
from .flags import *


class Chain:
    """The Chain object represent a single chain variable fragment (scFv) antibody.

    A scFv can be part of either the heavy or light chain of an antibody.
    The nature of the chain is determined by querying the sequence
    to the Abnum server, and is implemented with the Chain.ab_numbering() method.

    Attributes:
        numbering (list): the name of each position occupied by amino acids in sequence
        mw (float): the cached molecular weight
        pI (float): the cached isoelectric point of the sequence
        cdr (tuple): tuple with two dictionaries for CDR and FR with the index of the amino acids
        in each region
        germline_identity (dict):
    """

    def __init__(self, sequence, name='Chain1', numbering_scheme=NUMBERING_FLAGS.CHOTHIA):
        """
        The default Chain object constructor, which required a string
        representing a scFv sequence.

        Args:
            sequence (str): Amino acid sequence
            name (name): Name of sequence
            numbering_scheme (str): numbering scheme name to perform alignment

        Examples:
            Instantiate a Chain object with the default constructor
            >>> from abpytools.core import Chain
            >>> from abpytools.core.flags import *
            >>> chain = Chain(sequence='MYSEQUENCE', name='my_seq', numbering_scheme=NUMBERING_FLAGS.CHOTHIA)
        """
        self._raw_sequence = sequence.upper()
        self._sequence = self._raw_sequence.replace('-', '')
        self._aligned_sequence = None
        self._name = name
        self._chain = None
        self.numbering = None
        self.hydrophobicity_matrix = None
        self.mw = None
        self.pI = None
        self.cdr = None
        self._numbering_scheme = numbering_scheme
        self._loading_status = 'Not Loaded'
        self.germline_identity = dict()
        self.germline = tuple()
        self._cache = Cache(max_cache_size=10)

    @classmethod
    def load_from_string(cls, sequence, name='Chain1', numbering_scheme=NUMBERING_FLAGS.CHOTHIA):
        """
        Returns an instantiated Chain object from a sequence
        Args:
            sequence:
            name:
            numbering_scheme:

        Returns:

        """
        new_chain = cls(sequence=sequence, name=name, numbering_scheme=numbering_scheme)
        new_chain.load()
        return new_chain

    def load(self):
        """
        Generates all the data:
        - Chain Numbering
        - Hydrophobicity matrix
        - Molecular weight
        - pI

        All the data is then stored in its respective attributes

        :return:

        """

        if self._loading_status in [NUMBERING_FLAGS.FAILED, NUMBERING_FLAGS.NOT_LOADED]:
            try:
                self.numbering = self.ab_numbering()
                self._loading_status = NUMBERING_FLAGS.LOADED
                self.load()

            except ValueError:
                self._loading_status = NUMBERING_FLAGS.FAILED

            except NumberingException:
                self._loading_status = NUMBERING_FLAGS.UNNUMBERED

        elif self._loading_status == NUMBERING_FLAGS.LOADED:
            self.hydrophobicity_matrix = self.ab_hydrophobicity_matrix()
            self.mw = self.ab_molecular_weight()
            self.pI = self.ab_pi()
            self.cdr = self.ab_regions()

        else:
            # this should never happen...
            raise ValueError("Unknown loading status")  # pragma: no cover

    @staticmethod
    def determine_chain_type(numbering):
        if numbering[0][0] == 'H':
            chain = CHAIN_FLAGS.HEAVY_CHAIN
        elif numbering[0][0] == 'L':
            chain = CHAIN_FLAGS.LIGHT_CHAIN
        else:
            # couldn't determine chain type
            chain = CHAIN_FLAGS.UNKNOWN_CHAIN  # pragma: no cover

        return chain

    def ab_numbering(self, server=OPTION_FLAGS.ABYSIS, **kwargs):
        """
        Return list

        Returns:
            list:
        """
        # store the amino positions/numbering in a list -> len(numbering) == len(self._sequence)
        numbering = get_ab_numbering(self._sequence, server, self._numbering_scheme, **kwargs)
        self._chain = self.determine_chain_type(numbering)

        return numbering

    def ab_numbering_table(self, as_array=False, replacement='-', region='all'):

        """

        :param region:
        :param as_array: if True returns numpy.array object, if False returns a pandas.DataFrame
        :param replacement: value to replace empty positions
        :return:
        """

        region = numbering_table_region(region=region)

        # if the object has not been loaded successfully yet need to try and get the numbering scheme using
        # ab_numbering method
        if self._loading_status in [NUMBERING_FLAGS.NOT_LOADED,
                                    NUMBERING_FLAGS.FAILED]:
            self.numbering = self.ab_numbering()

        whole_sequence_dict, whole_sequence = numbering_table_sequences(region=region,
                                                                        numbering_scheme=self._numbering_scheme,
                                                                        chain=self._chain)

        # now that all the prep has been done we can extract the position from each amino acid
        # according to the numbering scheme
        data = np.empty((len(whole_sequence)), dtype=str)
        for i, position in enumerate(whole_sequence):
            if position in self.numbering:
                data[i] = self._sequence[self.numbering.index(position)]
            else:
                # if there is no amino acid in the sequence that corresponds to position i we just replace it by
                # the replacement value, which is by default '-'
                data[i] = replacement

        self._aligned_sequence = data

        if as_array:
            # return the data as numpy.array -> much faster than creating a pandas.DataFrame
            return data.reshape(1, -1)
        else:

            # return the data as a pandas.DataFrame -> it's slower but looks nicer and makes it easier to get
            # the data of interest

            multi_index = numbering_table_multiindex(region=region,
                                                     whole_sequence_dict=whole_sequence_dict)

            # create the DataFrame and assign the columns and index names
            data = pd.DataFrame(data=data).T
            data.columns = multi_index
            data.index = [self._name]
            return data

    def ab_hydrophobicity_matrix(self, hydrophobicity_scores=HYDROPHOBICITY_FLAGS.EW):

        # check if all the required parameters are in order
        if isinstance(hydrophobicity_scores, str):
            if hydrophobicity_scores not in OPTION_FLAGS.AVAILABLE_HYDROPHOBITY_SCORES:
                raise ValueError("Chosen hydrophobicity scores ({}) not available. \
                Available hydrophobicity scores: {}".format(
                    hydrophobicity_scores, ' ,'.join(OPTION_FLAGS.AVAILABLE_HYDROPHOBITY_SCORES)
                ))

        if self._loading_status == NUMBERING_FLAGS.NOT_LOADED:
            self.numbering = self.ab_numbering()
        if self._chain == 'NA':
            raise ValueError("Could not determine chain type")

        data_loader = DataLoader(data_type='NumberingSchemes',
                                 data=[self._numbering_scheme, self._chain])
        whole_sequence_dict = data_loader.get_data()

        # whole_sequence is a list with all the amino acid positions in the selected numbering scheme
        whole_sequence = whole_sequence_dict

        # get the dictionary with the hydrophobicity scores
        data_loader = DataLoader(data_type='AminoAcidProperties',
                                 data=['hydrophobicity', hydrophobicity_scores + 'Hydrophobicity'])
        aa_hydrophobicity_scores = data_loader.get_data()

        return calculate_hydrophobicity_matrix(whole_sequence=whole_sequence,
                                               numbering=self.numbering,
                                               aa_hydrophobicity_scores=aa_hydrophobicity_scores,
                                               sequence=self._sequence)

    def ab_regions(self):

        """
        method to determine Chain regions (CDR and Framework) of each amino acid in sequence

        :return:
        """

        if 'cdrs' not in self._cache:

            if self._loading_status == NUMBERING_FLAGS.NOT_LOADED:
                self.numbering, self._chain = self.ab_numbering()

            if self.numbering == 'NA':
                raise ValueError("Cannot return CDR positions without the antibody numbering information")

            data_loader = DataLoader(data_type='CDR_positions', data=[self._numbering_scheme, self._chain])
            cdr_positions = data_loader.get_data()

            data_loader = DataLoader(data_type='Framework_positions', data=[self._numbering_scheme, self._chain])
            framework_position = data_loader.get_data()

            cdrs = calculate_cdr(numbering=self.numbering, cdr_positions=cdr_positions,
                                 framework_positions=framework_position)

            self._cache.update('cdrs', cdrs)

        return self._cache['cdrs']

    def ab_molecular_weight(self, monoisotopic=False):

        if monoisotopic:
            data_loader = DataLoader(data_type='AminoAcidProperties',
                                     data=['MolecularWeight', 'average'])
        else:
            data_loader = DataLoader(data_type='AminoAcidProperties',
                                     data=['MolecularWeight', 'monoisotopic'])
        mw_dict = data_loader.get_data()

        return calculate_mw(self._sequence, mw_dict)

    def ab_pi(self, pi_database='Wikipedia'):

        assert pi_database in OPTION_FLAGS.AVAILABLE_PI_VALUES, \
            "Selected pI database {} not available. " \
            "Available databases: {}".format(pi_database, ' ,'.join(
                OPTION_FLAGS.AVAILABLE_PI_VALUES))

        data_loader = DataLoader(data_type='AminoAcidProperties',
                                 data=['pI', pi_database])
        pi_data = data_loader.get_data()

        return calculate_pi(sequence=self._sequence, pi_data=pi_data)

    def ab_ec(self, extinction_coefficient_database='Standard', reduced=False, normalise=False, **kwargs):

        if reduced:
            extinction_coefficient_database += '_reduced'

        data_loader = DataLoader(data_type='AminoAcidProperties', data=['ExtinctionCoefficient',
                                                                        extinction_coefficient_database])

        ec_data = data_loader.get_data()

        if normalise:
            return calculate_ec(sequence=self._sequence, ec_data=ec_data) / self.ab_molecular_weight(**kwargs)
        else:
            return calculate_ec(sequence=self._sequence, ec_data=ec_data)

    def ab_format(self):
        return {"name": self._name, "sequence": self._sequence, "numbering": self.numbering, "chain": self._chain,
                "MW": self.mw, "CDR": self.cdr, "numbering_scheme": self._numbering_scheme, "pI": self.pI}

    def ab_charge(self, align=True, ph=7.4, pka_database='Wikipedia'):

        """
        Method to calculate the charges for each amino acid of antibody
        :param pka_database:
        :param ph:
        :param align: if set to True an alignment will be performed,
                      if it hasn't been done already using the ab_numbering method

        :return: array with amino acid charges
        """

        assert pka_database in OPTION_FLAGS.AVAILABLE_PI_VALUES, \
            "Selected pI database {} not available. " \
            "Available databases: {}".format(pka_database,
                                             ' ,'.join(OPTION_FLAGS.AVAILABLE_PI_VALUES))

        data_loader = DataLoader(data_type='AminoAcidProperties',
                                 data=['pI', pka_database])
        pka_data = data_loader.get_data()

        if align:
            # get the first (and only) row
            sequence = self.ab_numbering_table(as_array=True)[0]
        else:
            sequence = list(self.sequence)

        return np.array([amino_acid_charge(x, ph, pka_data) for x in sequence])

    def ab_total_charge(self, ph=7.4, pka_database=PI_FLAGS.WIKIPEDIA):

        assert pka_database in OPTION_FLAGS.AVAILABLE_PI_VALUES, \
            "Selected pI database {} not available. " \
            "Available databases: {}".format(pka_database,
                                             ' ,'.join(OPTION_FLAGS.AVAILABLE_PI_VALUES))

        data_loader = DataLoader(data_type='AminoAcidProperties',
                                 data=['pI', pka_database])
        pka_data = data_loader.get_data()

        return calculate_charge(sequence=self._sequence, ph=ph, pka_values=pka_data)

    @property
    def chain(self):
        return self._chain

    @property
    def name(self):
        return self._name

    def set_name(self, name):
        self._name = name

    @property
    def sequence(self):
        return self._sequence

    @property
    def aligned_sequence(self):
        if self._aligned_sequence is None:
            _ = self.ab_numbering_table(as_array=True, replacement='-')
        return self._aligned_sequence.tolist()

    @property
    def status(self):
        return self._loading_status

    @property
    def numbering_scheme(self):
        return self._numbering_scheme

    def _string_summary_basic(self):
        return "abpytools.Chain Name: {}, Chain type: {}, Sequence length: {}, Status: {}".format(
            self._name, self._chain, len(self._sequence),
            self._loading_status)  # pragma: no cover

    def __repr__(self):
        return "<%s at 0x%02x>" % (self._string_summary_basic(), id(self))  # pragma: no cover

    def __len__(self):
        return len(self.sequence)


def get_ab_numbering(sequence, server, numbering_scheme, timeout=30):
    """

    :rtype: list
    """
    # check which server to use to get numbering
    if server == OPTION_FLAGS.ABYSIS:
        # find out which numbering scheme to use
        if numbering_scheme == NUMBERING_FLAGS.CHOTHIA:
            scheme = '-c'
        elif numbering_scheme in (NUMBERING_FLAGS.CHOTHIA_EXT, NUMBERING_FLAGS.MARTIN):
            scheme = '-a'
        elif numbering_scheme == NUMBERING_FLAGS.KABAT:
            scheme = '-k'
        else:
            raise ValueError("{} numbering scheme is unknown.".format(numbering_scheme.capitalize()))

        # prepare the url string to query server
        url = f"http://www.bioinf.org.uk/cgi-bin/abnum/abnum.pl?plain=1&aaseq={sequence}&scheme={scheme}"
        # use the Download class from utils to get output
        numbering_table = Download(url, verbose=False, timeout=timeout)
        try:
            numbering_table.download()
        except ValueError:
            raise ValueError("Check the internet connection.")

        # check whether the server returned an error
        if numbering_table.html.replace("\n", '') == 'Warning: Unable to number sequence' or len(
                numbering_table.html.replace("\n", '')) == 0:
            raise NumberingException("Unable to number sequence")

        # parse the results
        parsed_numbering_table = re.findall("[\S| ]+", numbering_table.html)

        # get the numbering from the parsed table
        numbering = [x[:-2] for x in parsed_numbering_table if x[-1] != '-']

        # TODO: add more server options
    else:
        numbering = ['']

    return numbering


def calculate_hydrophobicity_matrix(whole_sequence, numbering, aa_hydrophobicity_scores, sequence):
    # instantiate numpy array (whole sequence includes all the amino acid positions of the VH/VL, even the ones
    # that aren't occupied -> these will be filled with zeros
    # hydrophobicity_matrix = np.zeros(len(whole_sequence))
    #
    # # iterate through each position
    # for i, position in enumerate(whole_sequence):
    #
    #     if position in numbering:
    #         position_in_data = numbering.index(position)
    #         hydrophobicity_matrix[i] = aa_hydrophobicity_scores[sequence[position_in_data]]
    #
    # return hydrophobicity_matrix
    # same thing as above but in a comprehension list
    return np.array([aa_hydrophobicity_scores[sequence[numbering.index(x)]] if x in numbering
                     else 0 for x in whole_sequence])


def calculate_mw(sequence, mw_data):
    return sum(mw_data[x] for x in sequence) - (len(sequence) - 1) * mw_data['water']


def calculate_ec(sequence, ec_data):
    # ϵ280 = nW x 5,500 + nY x 1,490 + nC x 125
    n_W = sequence.count('W')
    n_Y = sequence.count('Y')
    n_C = sequence.count('C')
    return n_W * ec_data['W'] + n_Y * ec_data['Y'] + n_C * ec_data['C']


def calculate_pi(sequence, pi_data):
    # algorithm implemented from http://isoelectric.ovh.org/files/practise-isoelectric-point.html

    # count number of D, E, C, Y, H, K, R
    d_count = sequence.count('D')
    e_count = sequence.count('E')
    c_count = sequence.count('C')
    y_count = sequence.count('Y')
    h_count = sequence.count('H')
    k_count = sequence.count('K')
    r_count = sequence.count('R')

    # initiate value of pH and nq (any number above 0)
    nq = 10
    ph = 0
    # define precision
    delta = 0.01

    while nq > 0:

        if ph >= 14:
            raise Exception("Could not calculate pI (pH reached above 14)")

        # qn1, qn2, qn3, qn4, qn5, qp1, qp2, qp3, qp4
        qn1 = -1 / (1 + 10 ** (pi_data['COOH'] - ph))  # C-terminus charge
        qn2 = - d_count / (1 + 10 ** (pi_data['D'] - ph))  # D charge
        qn3 = - e_count / (1 + 10 ** (pi_data['E'] - ph))  # E charge
        qn4 = - c_count / (1 + 10 ** (pi_data['C'] - ph))  # C charge
        qn5 = - y_count / (1 + 10 ** (pi_data['Y'] - ph))  # Y charge
        qp1 = h_count / (1 + 10 ** (ph - pi_data['H']))  # H charge
        qp2 = 1 / (1 + 10 ** (ph - pi_data['NH2']))  # N-terminus charge
        qp3 = k_count / (1 + 10 ** (ph - pi_data['K']))  # K charge
        qp4 = r_count / (1 + 10 ** (ph - pi_data['R']))  # R charge

        nq = qn1 + qn2 + qn3 + qn4 + qn5 + qp1 + qp2 + qp3 + qp4

        # update pH
        ph += delta

    return ph


def calculate_cdr(numbering, cdr_positions, framework_positions):
    """

    :param numbering:
    :param cdr_positions:
    :param framework_positions:
    :return:
    """

    cdrs = {'CDR1': list(),
            'CDR2': list(),
            'CDR3': list()}

    frameworks = {'FR1': list(),
                  'FR2': list(),
                  'FR3': list(),
                  'FR4': list()}

    for cdr in cdrs.keys():

        cdr_positions_i = cdr_positions[cdr]

        for i, position in enumerate(numbering):
            if position in cdr_positions_i:
                cdrs[cdr].append(i)

    for framework in frameworks.keys():

        framework_position_i = framework_positions[framework]

        for i, position in enumerate(numbering):
            if position in framework_position_i:
                frameworks[framework].append(i)

    return cdrs, frameworks


def amino_acid_charge(amino_acid, ph, pka_values):
    if amino_acid in ['D', 'E', 'C', 'Y']:
        return -1 / (1 + 10 ** (pka_values[amino_acid] - ph))
    elif amino_acid in ['K', 'R', 'H']:
        return 1 / (1 + 10 ** (ph - pka_values[amino_acid]))
    else:
        return 0


def calculate_charge(sequence, ph, pka_values):
    # This calculation would make more sense but is slower (~1.5-2x)
    # cooh = -1 / (1 + 10 ** (pka_values['COOH'] - ph))
    # nh2 = 1 / (1 + 10 ** (ph - pka_values['NH2']))
    #
    # return sum([amino_acid_charge(x, ph, pka_values) for x in list(sequence)]) + cooh + nh2

    # Faster implementation
    # count number of D, E, C, Y, H, K, R
    d_count = sequence.count('D')
    e_count = sequence.count('E')
    c_count = sequence.count('C')
    y_count = sequence.count('Y')
    h_count = sequence.count('H')
    k_count = sequence.count('K')
    r_count = sequence.count('R')

    # qn1, qn2, qn3, qn4, qn5, qp1, qp2, qp3, qp4
    qn1 = -1 / (1 + 10 ** (pka_values['COOH'] - ph))  # C-terminus charge
    qn2 = - d_count / (1 + 10 ** (pka_values['D'] - ph))  # D charge
    qn3 = - e_count / (1 + 10 ** (pka_values['E'] - ph))  # E charge
    qn4 = - c_count / (1 + 10 ** (pka_values['C'] - ph))  # C charge
    qn5 = - y_count / (1 + 10 ** (pka_values['Y'] - ph))  # Y charge
    qp1 = h_count / (1 + 10 ** (ph - pka_values['H']))  # H charge
    qp2 = 1 / (1 + 10 ** (ph - pka_values['NH2']))  # N-terminus charge
    qp3 = k_count / (1 + 10 ** (ph - pka_values['K']))  # K charge
    qp4 = r_count / (1 + 10 ** (ph - pka_values['R']))  # R charge

    nq = qn1 + qn2 + qn3 + qn4 + qn5 + qp1 + qp2 + qp3 + qp4

    return nq



from .numbering import *
from .hydrophobicity import *
from .pi import *

AVAILABLE_NUMBERING_SCHEMES = [CHOTHIA, KABAT, CHOTHIA_EXT, MARTIN]
AVAILABLE_SERVERS = [ABYSIS]
AVAILABLE_HYDROPHOBITY_SCORES = [KD, WW, HH, MF, EW]
# to not be confused with math pi (this is pI)
AVAILABLE_PI_VALUES = [EMBOSS, DTASELECT, SOLOMON, SILLERO, RODWELL, WIKIPEDIA, LEHNINGER, GRIMSLEY]



class Cache:
    def __init__(self, max_cache_size=10):
        self.max_cache_size = max_cache_size
        self.cache = {}
        self._cache_size = 0
        self._cache_keys = []

    def update(self, key, data, override=True):
        # make space if the cache is full and override is True
        if len(self) >= self.max_cache_size and override:
            self._clear_cache()

        if len(self) >= self.max_cache_size and not override:
            raise ValueError("Cache is full, either increase the size of cache, "
                             "remove items from cache or allow override")

        if key not in self:
            self.add(key, data)

    def _clear_cache(self):
        while len(self) >= self.max_cache_size:
            # clear up cache until reaching max_cache_size - 1
            self.remove(self._cache_keys[0])

    def remove(self, key):
        self.cache.pop(key)
        self._cache_keys.pop(self._cache_keys.index(key))
        self._cache_size -= 1

    def add(self, key, data):
        self.cache[key] = data
        self._cache_keys.append(key)
        self._cache_size += 1

    def empty_cache(self):
        self.cache = {}
        self._cache_keys = []
        self._cache_size = 0

    def _string_summary_basic(self):
        return "abpytools.Cache size: {}".format(len(self))

    def __getitem__(self, item):
        return self.cache[item]

    def __contains__(self, item):
        if item in self._cache_keys:
            return True
        else:
            return False

    def __len__(self):
        return self._cache_size

    def __repr__(self):
        return "<%s at 0x%02x>" % (self._string_summary_basic(), id(self))

    def __setitem__(self, key, value):
        self.update(key, value)



from abpytools.core.chain_collection import ChainCollection
from abpytools.core.cache import Cache
import numpy as np


class ChainDomains(ChainCollection):

    def __init__(self, antibody_objects=None, path=None, verbose=True, show_progressbar=True, n_threads=10):
        super().__init__(antibody_objects=antibody_objects)
        if antibody_objects:
            self.load()
        else:
            self.__init__(antibody_objects=ChainCollection.load_from_file(path=path, verbose=verbose,
                                                                          show_progressbar=show_progressbar,
                                                                          n_threads=n_threads))

        self._cache = Cache(max_cache_size=5)

    def cdr_lengths(self):
        """
        method to obtain cdr_lengths
        :return: m by n matrix with CDR lengths, where m is the number of antibodies in ChainCollection and n is
        three, corresponding to the three CDRs.
        """

        if 'cdr_lengths' not in self._cache:

            cdr_length_matrix = np.zeros((self.n_ab, 3), dtype=np.int)
            cdr_sequences = self.cdr_sequences()

            for m, antibody in enumerate(self.antibody_objects):
                for n, cdr in enumerate(['CDR1', 'CDR2', 'CDR3']):
                    cdr_length_matrix[m, n] = len(cdr_sequences[antibody.name][cdr])

            self._cache.update(key='cdr_lengths', data=cdr_length_matrix)

        return self._cache['cdr_lengths']

    def cdr_sequences(self):
        """
        method that returns sequences of each cdr
        :return: list of dictionaries with keys 'CDR1', 'CDR2' and 'CDR3' containing a string with the respective amino
        acid sequence
        """

        if 'cdr_sequences' not in self._cache:

            cdr_sequences = dict()

            for antibody in self.antibody_objects:
                dict_i = dict()
                for cdr in ['CDR1', 'CDR2', 'CDR3']:
                    self.sequence_splitter_helper(antibody=antibody,
                                                  region=cdr,
                                                  index=0,
                                                  dict_i=dict_i)
                cdr_sequences[antibody.name] = dict_i

            self._cache.update(key='cdr_sequences', data=cdr_sequences)

        return self._cache['cdr_sequences']

    def framework_length(self):
        framework_length_matrix = np.zeros((self.n_ab, 4), dtype=np.int)
        fr_sequences = self.framework_sequences()
        for m, antibody in enumerate(self.antibody_objects):
            for n, framework in enumerate(['FR1', 'FR2', 'FR3', 'FR4']):
                framework_length_matrix[m, n] = len(fr_sequences[antibody.name][framework])

        return framework_length_matrix

    def framework_sequences(self):

        framework_sequences = dict()

        for antibody in self.antibody_objects:
            dict_i = dict()
            for framework in ['FR1', 'FR2', 'FR3', 'FR4']:
                self.sequence_splitter_helper(antibody=antibody,
                                              region=framework,
                                              index=1,
                                              dict_i=dict_i)
            framework_sequences[antibody.name] = dict_i

        return framework_sequences

    @staticmethod
    def sequence_splitter_helper(antibody, region, index, dict_i):
        seq_i = list()
        indices = antibody.ab_regions()[index][region]
        for i in indices:
            seq_i.append(antibody.sequence[i])
        dict_i[region] = ''.join(seq_i)



from abpytools.cython_extensions.convert_py_2_C cimport release_C_pointer, get_C_char_array_pointers
from abpytools.utils.math_utils cimport Matrix2D_backend, Vector
from libc.math cimport acos as acos_C
from libc.math cimport fmin as min_C
from libc.float cimport DBL_EPSILON


cdef Matrix2D_backend init_score_matrix_(str seq_1, str seq_2, int indel):
    """
    - score matrix initialisation with two sequences
    Example init_score_matrix('SEND', 'AND', -1):
        [[0, -1, -2],
         [-1, 0, 0],
         [-2, 0, 0],
         [-3, 0, 0]]

    Args:
        seq_1: 
        seq_2: 
        indel: 

    Returns:

    """

    init_matrix = Matrix2D_backend([[x * indel] + [0] * len(seq_1) if x > 0 else
                                    list(range(0, (len(seq_1) + 1) * indel, indel)) for x in range(len(seq_2) + 1)])

    return init_matrix


cpdef double cosine_distance_(list u, list v):
    """
    
    Args:
        u: 
        v: 

    Returns:

    """
    cdef int size
    if len(u) == len(v):
        size = len(u)
    else:
        raise ValueError("Vector size mismatch")

    cdef Vector u_ = Vector(u)
    cdef Vector v_ = Vector(v)
    cdef double result
    cdef double lower
    cdef int norm = 2

    cdef double upper = u_.dot_product(v_)

    if upper < DBL_EPSILON:
        # if the result is lower than the double rounding error just assume that it should be in fact 0
        result = 0

    else:
        lower = u_.norm(norm) * v_.norm(norm)
        result = acos_C(upper / lower)

    return result

cpdef int hamming_distance_(str seq1, str seq2):
    """
    
    Args:
        seq1: 
        seq2: 

    Returns:

    """

    cdef int size
    if len(seq1) == len(seq2):
        size = len(seq1)
    else:
        raise ValueError("Sequence size mismatch")

    cdef int i
    cdef int result = 0
    cdef char *seq1_ = get_C_char_array_pointers(seq1, size)
    cdef char *seq2_ = get_C_char_array_pointers(seq2, size)

    for i in range(size):
        if seq1_[i] != seq2_[i]:
            result+=1

    release_C_pointer(seq1_)
    release_C_pointer(seq2_)

    return result


cpdef double levenshtein_distance_(seq1, seq2):
    """
    
    Args:
        seq1: 
        seq2: 

    Returns:

    """
    cdef Matrix2D_backend dist = init_score_matrix_(seq_1=seq1, seq_2=seq2, indel=1)
    cdef int rows = dist.n_rows
    cdef int cols = dist.n_cols
    cdef int col, row, cost

    for col in range(1, cols):
        for row in range(1, rows):
            if seq2[row - 1] == seq1[col - 1]:
                cost = 0
            else:
                cost = 1
            dist[row, col] = min_C(min_C(dist[row - 1, col] + 1,  # deletion
                                   dist[row, col - 1] + 1),  # insertion
                                   dist[row - 1, col - 1] + cost)  # substitution

    return dist[rows-1, cols-1]



from .cache import Cache
from .base import CollectionBase
from .chain import Chain
from .chain_collection import ChainCollection
from .fab_collection import FabCollection
from .fab import Fab

__all__ = ["Chain", "ChainCollection", "FabCollection", "Fab"]



from ..utils import DataLoader
import itertools
import pandas as pd
import numpy as np


available_regions = ['FR1', 'CDR1', 'FR2', 'CDR2', 'FR3', 'CDR3', 'FR4']


def numbering_table_sequences(region, numbering_scheme, chain):
    # next two conditionals are used to only extract the needed data with FR and CDR positions
    # it makes the loading of the data quicker when there are only CDRs or FRs
    if any([True if x.startswith('CDR') else False for x in region]):
        cdr_list = DataLoader(data_type='CDR_positions',
                              data=[numbering_scheme, chain])
        cdrs = cdr_list.get_data()
    else:
        cdrs = {}

    if any([True if x.startswith('FR') else False for x in region]):
        fr_list = DataLoader(data_type='Framework_positions',
                             data=[numbering_scheme, chain])
        frs = fr_list.get_data()
    else:
        frs = {}

    # pack it all up into a single dictionary
    whole_sequence_dict = {**cdrs, **frs}
    # get the sequence list in the correct order (since region has been sorted before)
    whole_sequence_list = [whole_sequence_dict[x] for x in region]
    # unpack whole_sequence_list into a single list
    whole_sequence = list(itertools.chain.from_iterable(whole_sequence_list))

    return whole_sequence_dict, whole_sequence


def numbering_table_region(region):

    # if 'all' is chosen then region becomes a list with all
    if region == 'all':
        region = ['FR1', 'CDR1', 'FR2', 'CDR2', 'FR3', 'CDR3', 'FR4']

    # if region is a string (i.e. 'CDR1') it becomes a list and in the next block of code is checked if it
    # is a valid selection
    if isinstance(region, str):
        region = [region]

    # checks if the selected regions are all in the available_regions
    if not set(region).issubset(set(available_regions)):
        raise ValueError("The chosen region is not available"
                         "Currently available regions: {}".format(available_regions))

    # make sure the regions are in a logical order
    region.sort(key=lambda x: available_regions.index(x))

    return region


def numbering_table_multiindex(region, whole_sequence_dict):
    # [[(CDR1, L23), (CDR1, L24),..], ..., [(CDR3, L100), ...]]
    # here we get the list of lists from the line above
    pre_region_map = [[(region_i, numbering) for numbering in whole_sequence_dict[region_i]] for region_i in
                      region]

    # which can be easily unpacked into a single list
    # region_map is a list of tuples that can be interpreted by pd.MultiIndex to form a two layer column system
    region_map = list(itertools.chain.from_iterable(pre_region_map))
    multi_index = pd.MultiIndex.from_tuples(tuples=region_map, names=['Region', 'Numbering'])

    return multi_index


def germline_identity_pd(heavy_identity, light_identity, internal_heavy, internal_light, names):

    regions = ['CDR1', 'CDR2', 'CDR3', 'FR1', 'FR2', 'FR3', 'Total']

    columns = pd.MultiIndex.from_tuples([('Light', x) for x in regions] +
                                        [('Heavy', x) for x in regions] +
                                        [('Average', x) for x in regions],
                                        names=['Chain', 'Region'])

    df = pd.DataFrame(columns=columns, index=names)

    for column in columns:

        if column[0] == 'Light':
            df[column] = [light_identity[x][column[1]] if column[1] in light_identity[x] else np.NaN
                          for x in internal_light]
            # df[column] = list(map(lambda x: light_identity[x][column[1]] if column[1] in light_identity[x] else np.NaN,
            #                       internal_light))
        elif column[0] == 'Heavy':
            df[column] = [heavy_identity[x][column[1]] if column[1] in heavy_identity[x] else np.NaN
                          for x in internal_heavy]
            # df[column] = list(map(lambda x: heavy_identity[x][column[1]] if column[1] in heavy_identity[x] else np.NaN,
            #                       internal_heavy))
        else:
            df[column] = (df[('Light', column[1])] + df[('Heavy', column[1])]) / 2

    return df


def to_numbering_table(as_array, region, chain, heavy_chains_numbering_table,
                       light_chains_numbering_table, names, **kwargs):

    if chain == 'both':

        if as_array:
            t_heavy = heavy_chains_numbering_table(as_array=True, region=region, **kwargs)
            t_light = light_chains_numbering_table(as_array=True, region=region, **kwargs)

            data = np.concatenate((t_light, t_heavy), axis=1)

        else:
            t_heavy = heavy_chains_numbering_table(as_array=False, region=region, **kwargs)
            t_light = light_chains_numbering_table(as_array=False, region=region, **kwargs)

            t_heavy.reset_index(drop=True, inplace=True)
            t_light.reset_index(drop=True, inplace=True)

            data = pd.concat([t_light, t_heavy], axis=1, keys=['Light', 'Heavy'])

    elif chain == 'heavy':

        data = heavy_chains_numbering_table(as_array=as_array, region=region, **kwargs)

    elif chain == 'light':

        data = light_chains_numbering_table(as_array=as_array, region=region, **kwargs)

    else:
        raise ValueError("Unknown chain.")

    if not as_array:
        data.index = names

    return data



from sklearn import cluster, decomposition
import numpy as np
from abpytools import ChainCollection
from matplotlib import pyplot as plt


class Cluster:

    def __init__(self, antibodies, metric='hydrophobicity', clustering_method='kmeans', decomposition_method='PCA'):

        if isinstance(antibodies, ChainCollection):
            self.antibodies = antibodies
        elif isinstance(antibodies, str):
            self.antibodies = ChainCollection(path=antibodies)
        if self.antibodies.n_ab == 0:
            self.antibodies.load()

        self.metric = metric
        self.clustering_method = clustering_method
        self.decomposition_method = decomposition_method
        self.cluster_assignment = np.zeros(self.antibodies.n_ab, dtype=int)
        self.cluster_assignment_dict = dict()
        self._data = None

    def _collect_data(self):
        if self.metric == 'hydrophobicity':
            return self.antibodies.hydrophobicity_matrix()

    def cluster(self, n_components=0.95, n_clusters=3):

        if self.decomposition_method == 'PCA':
            decomposition_obj = decomposition.PCA(n_components)

        self._data = decomposition_obj.fit_transform(self._collect_data())

        if self.clustering_method == 'kmeans':
            clustering_obj = cluster.KMeans(n_clusters=n_clusters)

        self.cluster_assignment = clustering_obj.fit_predict(self._data)

        for i, antibody_obj in enumerate(self.antibodies.antibody_objects):

            assignment = 'Cluster_{}'.format(self.cluster_assignment[i])

            if assignment not in self.cluster_assignment_dict:
                self.cluster_assignment_dict[assignment] = list()

            self.cluster_assignment_dict[assignment].append(antibody_obj)

    def plot_cluster(self):

        if len(self.cluster_assignment_dict) == 0:
            self.cluster()

        color = iter(plt.get_cmap('Vega20').colors)

        plt.figure(figsize=(8, 8))

        for assignment in np.unique(self.cluster_assignment):
            c = next(color)

            plt.scatter(self._data[self.cluster_assignment == assignment, 0],
                        self._data[self.cluster_assignment == assignment, 1],
                        c=c, label='Cluster {}'.format(assignment))

            plt.legend(loc='best', prop={"size": 14})



from .chain_collection import ChainCollection
import numpy as np
import pandas as pd
from .chain import calculate_charge
from abpytools.utils import DataLoader
from operator import itemgetter
from .fab import Fab
from .helper_functions import germline_identity_pd, to_numbering_table
from .base import CollectionBase
import os
import json
from .utils import (json_FabCollection_formatter, pb2_FabCollection_formatter, pb2_FabCollection_parser,
                    json_FabCollection_parser)
from .flags import *

if BACKEND_FLAGS.HAS_PROTO:
    from abpytools.core.formats import FabCollectionProto


class FabCollection(CollectionBase):

    def __init__(self, fab=None, heavy_chains=None, light_chains=None, names=None):
        """
        Fab object container that handles combinations of light/heavy Chain pairs.

        Args:
            fab (list):
            heavy_chains (ChainCollection):
            light_chains (ChainCollection):
            names (list):
        """
        # check if it's a Chain object
        if heavy_chains is None and light_chains is None and fab is None:
            raise ValueError('Provide a list of Chain objects or an ChainCollection object')

        # check if fab object is a list and if all object are abpytools.Fab objects
        if isinstance(fab, list) and all(isinstance(fab_i, Fab) for fab_i in fab):
            self._fab = fab
            self._light_chains = ChainCollection([x[0] for x in self._fab])
            self._heavy_chains = ChainCollection([x[1] for x in self._fab])

        if fab is None and (heavy_chains is not None and light_chains is not None):

            if isinstance(heavy_chains, list):
                self._heavy_chains = ChainCollection(antibody_objects=heavy_chains)

            elif isinstance(heavy_chains, ChainCollection):
                self._heavy_chains = heavy_chains

            else:
                raise ValueError('Provide a list of Chain objects or an ChainCollection object')

            if isinstance(light_chains, list):
                self._light_chains = ChainCollection(antibody_objects=light_chains)

            elif isinstance(light_chains, ChainCollection):
                self._light_chains = light_chains

            else:
                raise ValueError('Provide a list of Chain objects or an ChainCollection object')

            if len(self._light_chains.loading_status()) == 0:
                self._light_chains.load()

            if len(self._heavy_chains.loading_status()) == 0:
                self._heavy_chains.load()

            if self._light_chains.n_ab != self._heavy_chains.n_ab:
                raise ValueError('Number of heavy chains must be the same of light chains')

        if isinstance(names, list) and all(isinstance(name, str) for name in names):
            if len(names) == self._heavy_chains.n_ab:
                self._names = names
            else:
                raise ValueError(
                    'Length of name list must be the same as length of heavy_chains/light chains lists')

        elif names is None:
            self._names = ['{} - {}'.format(heavy, light) for heavy, light in zip(self._heavy_chains.names,
                                                                                  self._light_chains.names)]

        else:
            raise ValueError("Names expected a list of strings, instead got {}".format(type(names)))

        self._n_ab = self._light_chains.n_ab
        self._pair_sequences = [heavy + light for light, heavy in zip(self._heavy_chains.sequences,
                                                                      self._light_chains.sequences)]

        # keep the name of the heavy and light chains internally to keep everything in the right order
        self._internal_heavy_name = self._heavy_chains.names
        self._internal_light_name = self._light_chains.names

    # even though it makes more sense to draw all these values from the base Fab objects this is much slower
    # whenever self._n_ab > 1 it makes more sense to use the self._heavy_chain and self._light_chain containers
    # in all the methods
    # in essence the abpytools.Fab object is just a representative building block that could in future just
    # cache data and would then represent a speed up in the calculations
    def molecular_weights(self, monoisotopic=False):

        return [heavy + light for heavy, light in zip(self._heavy_chains.molecular_weights(monoisotopic=monoisotopic),
                                                      self._light_chains.molecular_weights(monoisotopic=monoisotopic))]

    def extinction_coefficients(self, extinction_coefficient_database='Standard', reduced=False, normalise=False,
                                **kwargs):

        heavy_ec = self._heavy_chains.extinction_coefficients(
            extinction_coefficient_database=extinction_coefficient_database,
            reduced=reduced)
        light_ec = self._light_chains.extinction_coefficients(
            extinction_coefficient_database=extinction_coefficient_database,
            reduced=reduced)

        if normalise:
            return [(heavy + light) / mw for heavy, light, mw in
                    zip(heavy_ec, light_ec, self.molecular_weights(**kwargs))]
        else:
            return [heavy + light for heavy, light in zip(heavy_ec, light_ec)]

    def hydrophobicity_matrix(self):

        return np.column_stack((self._heavy_chains.hydrophobicity_matrix(), self._light_chains.hydrophobicity_matrix()))

    def charge(self):

        return np.column_stack((self._heavy_chains.charge, self._light_chains.charge))

    def total_charge(self, ph=7.4, pka_database='Wikipedia'):

        available_pi_databases = ["EMBOSS", "DTASetect", "Solomon", "Sillero", "Rodwell", "Wikipedia", "Lehninger",
                                  "Grimsley"]
        assert pka_database in available_pi_databases, \
            "Selected pI database {} not available. Available databases: {}".format(pka_database,
                                                                                    ' ,'.join(available_pi_databases))

        data_loader = DataLoader(data_type='AminoAcidProperties', data=['pI', pka_database])
        pka_data = data_loader.get_data()

        return [calculate_charge(sequence=seq, ph=ph, pka_values=pka_data) for seq in self.sequences]

    def igblast_local_query(self, file_path, chain):

        if chain.lower() == 'light':
            self._light_chains.igblast_local_query(file_path=file_path)
        elif chain.lower() == 'heavy':
            self._heavy_chains.igblast_local_query(file_path=file_path)
        else:
            raise ValueError('Specify if the data being loaded is for the heavy or light chain')

    def igblast_server_query(self, **kwargs):
        self._light_chains.igblast_server_query(**kwargs)
        self._heavy_chains.igblast_server_query(**kwargs)

    def numbering_table(self, as_array=False, region='all', chain='both', **kwargs):

        return to_numbering_table(as_array=as_array, region=region, chain=chain,
                                  heavy_chains_numbering_table=self._heavy_chains.numbering_table,
                                  light_chains_numbering_table=self._light_chains.numbering_table,
                                  names=self.names, **kwargs)

    def _germline_pd(self):

        # empty dictionaries return false, so this condition checks if any of the values are False
        if all([x for x in self._light_chains.germline_identity.values()]) is False:
            # this means there is no information about the germline,
            # by default it will run a web query
            self._light_chains.igblast_server_query()
        if all([x for x in self._heavy_chains.germline_identity.values()]) is False:
            self._heavy_chains.igblast_server_query()

        heavy_chain_germlines = self._heavy_chains.germline
        light_chain_germlines = self._light_chains.germline

        data = np.array([[heavy_chain_germlines[x][0] for x in self._internal_heavy_name],
                         [heavy_chain_germlines[x][1] for x in self._internal_heavy_name],
                         [light_chain_germlines[x][0] for x in self._internal_light_name],
                         [light_chain_germlines[x][1] for x in self._internal_light_name]]).T

        df = pd.DataFrame(data=data,
                          columns=pd.MultiIndex.from_tuples([('Heavy', 'Assignment'),
                                                             ('Heavy', 'Score'),
                                                             ('Light', 'Assignment'),
                                                             ('Light', 'Score')]),
                          index=self.names)

        df.loc[:, (slice(None), 'Score')] = df.loc[:, (slice(None), 'Score')].apply(pd.to_numeric)

        return df

    def save_to_json(self, path, update=True):
        with open(os.path.join(path + '.json'), 'w') as f:
            fab_data = json_FabCollection_formatter(self)
            json.dump(fab_data, f, indent=2)

    def save_to_pb2(self, path, update=True):
        proto_parser = FabCollectionProto()
        try:
            with open(os.path.join(path + '.pb2'), 'rb') as f:
                proto_parser.ParseFromString(f.read())
        except IOError:
            # Creating new file
            pass

        pb2_FabCollection_formatter(self, proto_parser)

        with open(os.path.join(path + '.pb2'), 'wb') as f:
            f.write(proto_parser.SerializeToString())

    def save_to_fasta(self, path, update=True):
        raise NotImplementedError

    @classmethod
    def load_from_json(cls, path, n_threads=20, verbose=True, show_progressbar=True):
        with open(path, 'r') as f:
            data = json.load(f)

        fab_objects = json_FabCollection_parser(data)

        fab_collection = cls(fab=fab_objects)

        return fab_collection

    @classmethod
    def load_from_pb2(cls, path, n_threads=20, verbose=True, show_progressbar=True):
        with open(path, 'rb') as f:
            proto_parser = FabCollectionProto()
            proto_parser.ParseFromString(f.read())

        fab_objects = pb2_FabCollection_parser(proto_parser)

        fab_collection = cls(fab=fab_objects)

        return fab_collection

    @classmethod
    def load_from_fasta(cls, path, numbering_scheme=NUMBERING_FLAGS.CHOTHIA, n_threads=20,
                        verbose=True, show_progressbar=True):
        raise NotImplementedError

    def _get_names_iter(self, chain='both'):
        if chain == 'both':
            for light_chain, heavy_chain in zip(self._light_chains, self._heavy_chains):
                yield f"{light_chain.name}-{heavy_chain.name}"
        elif chain == 'light':
            for light_chain in self._light_chains:
                yield light_chain.name
        elif chain == 'heavy':
            for heavy_chain in self._heavy_chains:
                yield heavy_chain.name
        else:
            raise ValueError(f"Unknown chain type ({chain}), available options are:"
                             f"both, light or heavy.")

    @property
    def regions(self):
        heavy_regions = self._heavy_chains.ab_region_index()
        light_regions = self._light_chains.ab_region_index()

        return {name: {CHAIN_FLAGS.HEAVY_CHAIN: heavy_regions[heavy],
                       CHAIN_FLAGS.LIGHT_CHAIN: light_regions[light]} for name, heavy, light in
                zip(self.names, self._internal_heavy_name, self._internal_light_name)}

    @property
    def names(self):
        return self._names

    @property
    def sequences(self):
        return self._pair_sequences

    @property
    def aligned_sequences(self):
        return [heavy + light for light, heavy in
                zip(self._heavy_chains.aligned_sequences,
                    self._light_chains.aligned_sequences)]

    @property
    def n_ab(self):
        return self._n_ab

    @property
    def germline_identity(self):
        return self._germline_identity()

    @property
    def germline(self):
        return self._germline_pd()

    def _string_summary_basic(self):
        return "abpytools.FabCollection Number of sequences: {}".format(self._n_ab)

    def __len__(self):
        return self._n_ab

    def __repr__(self):
        return "<%s at 0x%02x>" % (self._string_summary_basic(), id(self))

    def __getitem__(self, indices):
        if isinstance(indices, int):
            return Fab(heavy_chain=self._heavy_chains[indices],
                       light_chain=self._light_chains[indices],
                       name=self.names[indices], load=False)
        else:
            return FabCollection(heavy_chains=list(itemgetter(*indices)(self._heavy_chains)),
                                 light_chains=list(itemgetter(*indices)(self._light_chains)),
                                 names=list(itemgetter(*indices)(self._names)))

    def _germline_identity(self):

        # empty dictionaries return false, so this condition checks if any of the values are False
        if all([x for x in self._light_chains.germline_identity.values()]) is False:
            # this means there is no information about the germline,
            # by default it will run a web query
            self._light_chains.igblast_server_query()
        if all([x for x in self._heavy_chains.germline_identity.values()]) is False:
            self._heavy_chains.igblast_server_query()

        return germline_identity_pd(self._heavy_chains.germline_identity,
                                    self._light_chains.germline_identity,
                                    self._internal_heavy_name,
                                    self._internal_light_name,
                                    self._names)

    def get_object(self, name):

        """

        :param name: str
        :return:
        """

        if name in self.names:
            index = self.names.index(name)
            return self[index]
        else:
            raise ValueError('Could not find sequence with specified name')



import warnings
import _pickle as cPickle
from abpytools.home import Home
from ..utils.python_config import PythonConfig
import matplotlib.pyplot as plt

SUPPORTED_SUBSITUTION_MATRICES = ['BLOSUM45', 'BLOSUM62', 'BLOSUM80']


def load_alignment_algorithm(algorithm):
    if algorithm.lower() == 'needleman_wunsch':
        return needleman_wunsch

    else:
        raise ValueError("Unknown algorithm")


def load_substitution_matrix(substitution_matrix):

    abpytools_directory = Home().homedir

    if substitution_matrix in SUPPORTED_SUBSITUTION_MATRICES:
        with open('{}/data/{}.txt'.format(abpytools_directory, substitution_matrix), 'rb') as f:
            matrix = cPickle.load(f)
    else:
        raise ValueError("Unknown substitution matrix")

    return matrix


def needleman_wunsch(seq_1, seq_2, substitution_matrix, indel=-1):

    if indel >= 0:
        f = "Indel must be negative, setting indel to {}.".format(-indel)
        warnings.warn(f)
        indel = -indel

    # initialise matrix
    init_matrix = init_score_matrix(seq_1=seq_1, seq_2=seq_2,
                                    indel=indel)

    scores, traceback_matrix = calculate_scores(matrix=init_matrix,
                                                seq_1=seq_1,
                                                seq_2=seq_2,
                                                substitution_matrix=substitution_matrix,
                                                gap_penalty=indel)

    seq_2_aligned = traceback(traceback_matrix=traceback_matrix, seq_1=seq_1, seq_2=seq_2)

    return seq_2_aligned, scores[-1][-1]


def init_score_matrix(seq_1, seq_2, indel):
    """
    - score matrix initialisation with two sequences
    - pure python, i.e. no numpy
    Example init_score_matrix('SEND', 'AND', -1):
        [[0, -1, -2],
         [-1, 0, 0],
         [-2, 0, 0],
         [-3, 0, 0]]
    """

    init_matrix = [[x * indel] + [0] * len(seq_1) if x > 0 else list(range(0, (len(seq_1) + 1) * indel, indel)) for x in
                   range(len(seq_2) + 1)]

    return init_matrix


def calculate_scores(matrix, seq_1, seq_2, substitution_matrix, gap_penalty):
    traceback_matrix = [['up'] + [''] * len(seq_1) if x > 0 else ['left'] * (len(seq_1) + 1) for x in
                        range(len(seq_2) + 1)]
    traceback_matrix[0][0] = 'done'

    for i in range(1, len(matrix)):
        for j in range(1, len(matrix[i])):
            q_diag = matrix[i - 1][j - 1] + int(substitution_matrix[(seq_2[i - 1], seq_1[j - 1])])
            q_up = matrix[i - 1][j] + gap_penalty
            q_left = matrix[i][j - 1] + gap_penalty
            results = [q_diag, q_up, q_left]
            matrix[i][j] = max(results)
            max_index = results.index(matrix[i][j])
            if max_index == 0:
                traceback_matrix[i][j] = 'diag'
            elif max_index == 1:
                traceback_matrix[i][j] = 'up'
            else:
                traceback_matrix[i][j] = 'left'

    return matrix, traceback_matrix


def traceback(traceback_matrix, seq_1, seq_2):
    row = -1
    column = -1
    current = traceback_matrix[row][column]

    # iter_seq_1 = iter(seq_1[::-1])
    # iter_seq_2 = iter(seq_2[::-1])

    # seq_1 = seq_1[::-1]
    # seq_2 = seq_2[::-1]

    # aligned_seq_1 = list()
    aligned_seq_2 = list()

    while current != 'done':
        # is aligned
        if current == 'diag':
            # aligned_seq_1.append(next(iter_seq_1))
            # aligned_seq_2.append(next(iter_seq_2))
            # aligned_seq_1.append(seq_1[column])
            aligned_seq_2.append(seq_2[row])
            row -= 1
            column -= 1

        # leave a gap
        elif current == 'left':
            # aligned_seq_1.append(next(iter_seq_1))
            # aligned_seq_1.append(seq_1[column])
            aligned_seq_2.append('-')
            column -= 1

        # leave a gap
        else:
            # aligned_seq_1.append(next(iter_seq_1))
            # aligned_seq_1.append(seq_1[column])
            aligned_seq_2.append('-')
            row -= 1

        current = traceback_matrix[row][column]

    return ''.join(aligned_seq_2)[::-1]


def switch_interactive_mode(save=False):
    ipython_config = PythonConfig()
    if ipython_config.ipython_info == 'notebook' and save is False:
        if ipython_config.matplotlib_interactive is False:
            # turns on interactive mode
            plt.ion()



from abpytools.features.regions import ChainDomains
from collections import Counter
import numpy as np
from matplotlib import pyplot as plt
from abpytools.utils.data_loader import DataLoader
import os
from abpytools.utils import PythonConfig

amino_acid_index = {"R": 0,
                    "N": 1,
                    "D": 2,
                    "E": 3,
                    "Q": 4,
                    "K": 5,
                    "S": 6,
                    "T": 7,
                    "C": 8,
                    "H": 9,
                    "M": 10,
                    "A": 11,
                    "V": 12,
                    "G": 13,
                    "I": 14,
                    "L": 15,
                    "F": 16,
                    "P": 17,
                    "W": 18,
                    "Y": 19}


class AminoAcidFreq(ChainDomains):

    def __init__(self, antibody_objects=None, path=None, region='CDR3', load=False):
        super(AminoAcidFreq, self).__init__(antibody_objects=antibody_objects, path=path, load=load)

        regions = ['all', 'CDRs', 'FRs', 'FR1', 'FR2', 'FR3', 'FR4', 'CDR1', 'CDR2', 'CDR3']
        if region in regions:
            # get the sequence for the specified region
            self.region = region

            if self.region.startswith('CDR'):
                self._sequences = [self.cdr_sequences()[name][self.region] for name in self.names]
                data_loader = DataLoader(data_type='CDR_positions', data=['chothia', self.chain])
                self._numbering = data_loader.get_data()[self.region]

            elif self.region.startswith('FR'):
                self._sequences = [self.framework_sequences()[name][self.region] for name in self.names]

            # TODO: implement 'all'
            elif self.region == 'all':
                raise NotImplementedError("This is not the code you're looking for.")
        else:
            raise ValueError('Parameter region must be either: {}. Not {}'.format(' ,'.join(regions), region))

        self._sequence_count = len(max(self._sequences, key=len))

        self._aa_freq = np.zeros((20, self._sequence_count))
        self._aa_hyd_freq = np.zeros((3, self._sequence_count))
        self._aa_chg_freq = np.zeros((3, self._sequence_count))

        self._aa_count = np.zeros((20, self._sequence_count))
        self._aa_hyd_count = np.zeros((3, self._sequence_count))
        self._aa_chg_count = np.zeros((3, self._sequence_count))

    def _amino_acid_freq(self, normalize):

        # if the sum of self._aa_count is zero then the count has not been performed at this point
        if self._aa_count.sum() == 0:

            for position in range(len(max(self._sequences, key=len))):

                position_sequence = [x[position] for x in self._sequences if len(x) > position]
                count_i = Counter(position_sequence)
                total_i = len(position_sequence)

                for amino_acid_i in count_i.keys():

                    self._aa_count[amino_acid_index[amino_acid_i], position] = count_i[amino_acid_i]

                    # _aa_hyd_freq: row1 -> hydrophilic
                    #               row2 -> moderate
                    #               row3 -> hydrophobic
                    if amino_acid_i in ['R', 'N', 'D', 'E', 'Q', 'K', 'S', 'T']:
                        self._aa_hyd_count[0, position] += count_i[amino_acid_i]
                    elif amino_acid_i in ['C', 'H', 'M']:
                        self._aa_hyd_count[1, position] += count_i[amino_acid_i]
                    else:
                        self._aa_hyd_count[2, position] += count_i[amino_acid_i]

                    # _aa_chg_freq: row1 -> negative
                    #               row2 -> positive
                    #               row3 -> neutral
                    if amino_acid_i in ['D', 'E']:
                        self._aa_chg_count[0, position] += count_i[amino_acid_i]
                    elif amino_acid_i in ['R', 'K', 'H']:
                        self._aa_chg_count[1, position] += count_i[amino_acid_i]
                    else:
                        self._aa_chg_count[2, position] += count_i[amino_acid_i]

                # normalize values
                # doing it even when it is not required comes at a small computational cost
                # it would take longer if the user had to recalculate everything to have a count plot and then a
                # frequency plot
                self._aa_freq[:, position] = self._aa_count[:, position] / total_i
                self._aa_chg_freq[:, position] = self._aa_chg_count[:, position] / total_i
                self._aa_hyd_freq[:, position] = self._aa_hyd_count[:, position] / total_i

        if normalize:
            return self._aa_freq, self._aa_chg_freq, self._aa_hyd_freq
        else:
            return self._aa_count, self._aa_chg_count, self._aa_hyd_count

    def plot(self, sort_by='name', normalize=True, display_count=True, plot_path='./',
             plot_name='AminoAcidFrequency.png', notebook_plot=True):

        ipython_config = PythonConfig()
        if ipython_config.matplotlib_interactive is False and ipython_config.ipython_info == 'notebook':
            plt.ion()

        if sort_by not in ['name', 'hydropathy', 'charge']:
            raise ValueError("Argument for sort_by not valid. Valid arguments are name, hydrophobicity and charge")

        # get count/ freq matrices
        # to avoid writing more code than necessary the count and freq are stored in the same variable
        # since they will always be plotted independently
        aa, chg, hyd = self._amino_acid_freq(normalize=normalize)

        fig = plt.figure(1, figsize=(8, 8))
        ax = fig.add_subplot(111)

        for position in range(self._aa_freq.shape[1]):
            previous = 0
            # 20 distinct colors
            colors = ["#023fa5", "#7d87b9", "#bec1d4", "#d6bcc0", "#bb7784", "#8e063b", "#4a6fe3", "#8595e1",
                      "#b5bbe3", "#e6afb9", "#e07b91", "#d33f6a", "#11c638", "#8dd593", "#c6dec7", "#ead3c6",
                      "#f0b98d", "#ef9708", "#0fcfc0", "#9cded6"]

            if sort_by == 'name':

                # previous, lgd = self.plot_helper(ax=ax, colors=colors,
                #                                  title='amino acids',
                #                                  keys=sorted(amino_acid_index.keys()),
                #                                  position=position, data=aa,
                #                                  previous=previous)

                ax.set_title(self.region + ' amino acids', size=20)
                for i, amino_acid in enumerate(sorted(amino_acid_index.keys())):
                    c = colors[i]
                    ax.bar(position, aa[amino_acid_index[amino_acid], position], bottom=previous,
                           label=amino_acid, color=c, align='center')
                    previous += aa[amino_acid_index[amino_acid], position]
                lgd = ax.legend(sorted(amino_acid_index.keys()), loc='center left', bbox_to_anchor=(1, 0.5),
                                prop={"size": 16})

            elif sort_by == 'hydropathy':

                previous, lgd = self.plot_helper(ax=ax, colors=['b', 'r', 'k'],
                                                 title=sort_by,
                                                 keys=['Hydrophilic', 'Moderate',
                                                       'Hydrophobic'],
                                                 position=position, data=hyd,
                                                 previous=previous)

            else:

                previous, lgd = self.plot_helper(ax=ax, colors=['b', 'r', 'k'],
                                                 title='amino acid charge',
                                                 keys=['Negative', 'Positive', 'Neutral'],
                                                 position=position, data=chg,
                                                 previous=previous)

        if display_count:
            for position in range(aa.shape[1]):
                ax.text(x=position, y=aa[:, position].sum(),
                        s=str(int(self._aa_count[:, position].sum())),
                        rotation=45, ha='center', va='bottom')

        if normalize:
            ax.set_ylabel('Frequency', size=16)
        else:
            ax.set_ylabel('Count', size=16)

        ax.set_xticks(np.arange(len(self._numbering)))
        ax.set_xticklabels(self._numbering, rotation=60)
        ax.set_xlabel('Position', size=16)
        ax.set_ylim([0, aa.sum(0).max()*1.1])
        ax.margins(0.02)
        ax.grid(axis='y')

        if ipython_config.ipython_info == 'notebook' and notebook_plot:
            ax.plot()
        else:
            fig.savefig(os.path.join(plot_path, plot_name), bbox_extra_artists=(lgd,), bbox_inches='tight')
            plt.close(fig)

    def plot_helper(self, ax, colors, title, keys, position, data, previous):

        ax.set_title('{} {}'.format(self.region, title), size=20)
        for i, prop_i in enumerate(keys):
            c = colors[i]
            ax.bar(position, data[i, position], bottom=previous, label=prop_i, color=c, align='center')
            previous += data[i, position]
        lgd = ax.legend(keys, loc='center left', bbox_to_anchor=(1, 0.5),
                        prop={"size": 16})

        return previous, lgd



import sys
from matplotlib import pyplot as plt


class PythonConfig:
    def __init__(self):
        self._backend = get_ipython_info()
        self._matplotlib_interactive = plt.isinteractive()

    @property
    def ipython_info(self):
        return self._backend

    @property
    def matplotlib_interactive(self):
        return self._matplotlib_interactive  # pragma: no cover


def get_ipython_info():
    # code obtained from stackoverflow forum:
    # http://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
    ip = 'other'
    if 'ipykernel' in sys.modules:
        ip = 'notebook'
    elif 'IPython' in sys.modules:
        ip = 'terminal'
    return ip



from .chain_pb2 import Chain as ChainProto
from .chain_pb2 import ChainCollection as ChainCollectionProto
from .fab_pb2 import Fab as FabProto
from .fab_pb2 import FabCollection as FabCollectionProto



from .analysis_helper_functions import load_alignment_algorithm, load_substitution_matrix


class SequenceAlignment:

    """
    Sequence alignment with two chain objects   
    """

    def __init__(self, target, collection, algorithm, substitution_matrix):

        """
        
        :param chain_1: 
        :param collection: 
        :param algorithm: 
        :param substitution_matrix: 
        """

        self._algorithm = algorithm
        self._substitution_matrix = load_substitution_matrix(substitution_matrix)
        self.target = target
        self._collection = collection
        self._aligned_collection = dict()
        self._alignment_scores = dict()
        self._aligned = False

    def align_sequences(self, **kwargs):

        # perform the alignment for each chain object in collection and store results in dictionaries with keys
        # corresponding to names of the sequence to be aligned
        for seq in self._collection.antibody_objects:
            result = self._align(self.target, seq, **kwargs)
            self._aligned_collection[seq.name] = result[0]
            self._alignment_scores[seq.name] = result[1]

        self._aligned = True

    def print_aligned_sequences(self):

        if not self._aligned:
            raise ValueError("Method align_sequences must be called first to perform alignment.")

        final_string = self._aligned_sequences_string()

        print(*final_string, sep='\n')

    @property
    def target_sequence(self):
        return self.target.sequence

    @property
    def aligned_sequences(self):
        return self._aligned_collection

    @property
    def score(self):
        return self._alignment_scores

    def _align(self, seq_1, seq_2, **kwargs):

        # loads the function object that is then called
        self._algorithm_function = load_alignment_algorithm(self._algorithm)

        return self._algorithm_function(seq_1.sequence, seq_2.sequence,
                                        self._substitution_matrix, **kwargs)

    def _aligned_sequences_string(self):

        # find longest name of target sequence and aligned sequences (for display purposes)
        max_name = max(len(self.target.name), max(len(x) for x in self._collection.names))

        f = '{:>%d}: {}' % max_name
        f_score = '{:>%d}: {} (Score: {})' % max_name

        # store the final string in a list so that everything is printed at the end in one go
        final_string = list()

        final_string.append(f.format(self.target.name, self.target.sequence))
        final_string.append('-'*(len(self.target.name)+len(self.target.sequence)))

        for seq in self._collection.names:
            final_string.append(f_score.format(seq, self._aligned_collection[seq], self._alignment_scores[seq]))

        return final_string



from .distance_metrics_ import cosine_distance_, hamming_distance_, levenshtein_distance_
from abpytools.utils.math_utils import Vector
# from .analysis_helper_functions import init_score_matrix
# from math import acos
# from ..utils.math_utils import dot_product, magnitude


def cosine_distance(u, v):
    """
    returns the cosine distance between vectors u and v
    :param u:
    :param v:
    :return:
    """
    if u == v:
        return 0
    else:
        # pure python: acos(max(min(dot_product(u, v) / (magnitude(u) * magnitude(v)), 1), -1))
        return cosine_distance_(u, v)


def cosine_similarity(u, v):
    """
    returns the cosine similarity between vectors u and v
    :param u:
    :param v:
    :return:
    """
    return 1 - cosine_distance(u, v)


def hamming_distance(seq1, seq2):
    """
    returns the hamming distance between two sequences
    :param seq1:
    :param seq2:
    :return:
    """
    if len(seq1) != len(seq2):
        raise ValueError("Sequences must be equal length, instead got {} and {}".format(len(seq1), len(seq2)))
    # pure python:
    # sum(aa1 != aa2 for aa1, aa2 in zip(seq1, seq2))
    return hamming_distance_(seq1, seq2)


def levenshtein_distance(seq1, seq2):
    """

    :param seq1:
    :param seq2:
    :return:
    """

    # pure python:
    # dist = init_score_matrix(seq_1=seq1, seq_2=seq2, indel=1)
    #
    # cols = len(dist[0])
    # rows = len(dist)
    #
    # for col in range(1, cols):
    #     for row in range(1, rows):
    #         if seq2[row - 1] == seq1[col - 1]:
    #             cost = 0
    #         else:
    #             cost = 1
    #         dist[row][col] = min(dist[row - 1][col] + 1,  # deletion
    #                              dist[row][col - 1] + 1,  # insertion
    #                              dist[row - 1][col - 1] + cost)  # substitution
    #
    # return dist[rows-1][cols-1]
    return levenshtein_distance_(seq1, seq2)


def euclidean_distance(u, v):
    """
    returns the euclidean distance
    :param u:
    :param v:
    :return:
    """
    u = Vector(u)
    v = Vector(v)
    r = u - v
    return r.norm(2)


def manhattan_distance(u, v):
    """
    returns the Manhattan distance
    :param u:
    :param v:
    :return:
    """
    u = Vector(u)
    v = Vector(v)
    r = u - v

    return r.norm(1)


def norm(u, v, degree=2):

    u = Vector(u)
    v = Vector(v)
    r = u - v

    return r.norm(degree)



from libc.stdlib cimport malloc, free
from libc.string cimport strcpy
from cython.operator cimport dereference, postincrement
# from libc.stdio cimport printf


cdef double* get_C_double_array_pointers(list a, int size):

    """
    Internal function to convert a python list to C array
    :param a: 
    :param size: 
    :return: 
    """

    # allocate memory of a_ and b_ C arrays that will contain a copy of a and b lists, respectively
    IF SSE4_2:
        cdef double *a_ = <double *> memalign(16, size*sizeof(double))
    ELSE:
        cdef double *a_ = <double *> malloc(size*sizeof(double))


    if a_ is NULL:
        raise MemoryError("Failed to allocate memory!")

    cdef int i
    for i in range(size):
        a_[i] = a[i]
        i+=1

    return a_

cdef double** allocate_pp(int size):
    IF SSE4_2:
        cdef double **a_ = <double **> memalign(16, size*sizeof(double*))
    ELSE:
        cdef double **a_ = <double **> malloc(size*sizeof(double*))
    for i in range(size):
        a_[i] = <double *> malloc(sizeof(double))
    return a_

cdef double** get_C_double_array_pp(double* a, int size):

    """
    Internal function to convert a C array to a C array of pointers that point to pointers (double**)
    :param a: 
    :param size: 
    :return: 
    """

    # allocate memory of a_ C arrays that will contain a copy of list a
    IF SSE4_2:
        cdef double **a_ = <double **> memalign(16, size*sizeof(double*))
    ELSE:
        cdef double **a_ = <double **> malloc(size*sizeof(double*))

    if a_ is NULL:
        raise MemoryError("Failed to allocate memory!")

    for i in range(size):
        a_[i] = a
        # printf("%f, %p\n", dereference(a_[i]), a_[i]);
        # print(i, ref, dereference(a_[i]), hex(<unsigned long>&a_[i]))
        postincrement(a)

    return a_


cdef char* get_C_char_array_pointers(str a, int size):

    """
    Internal function to convert a python string to C array
    :param a: 
    :param size: 
    :return: 
    """

    # allocate memory of a_ and b_ C arrays that will contain a copy of a and b lists, respectively
    IF SSE4_2:
        cdef char *a_ = <char*> memalign(16, size*sizeof(char))
    ELSE:
        cdef char *a_ = <char*> malloc(size*sizeof(char))

    if a_ is NULL:
        raise MemoryError("Failed to allocate memory!")

    cdef bytes a_py_bytes = a.encode('UTF-8')

    strcpy(a_, a_py_bytes)

    return a_


cdef void release_C_pointer(scalar_or_char *a):

    """
    Internal function to release memory allocated for pointer *a.
    :param a: 
    :return: 
    """

    free(a)
    a=NULL


cdef void release_C_pp(scalar_or_char **a):

    free(a[0])
    free(a)
    a=NULL


cdef double** get_pp_from_ptr(double* ptr, int size):
    """
    Internal function to get C array from a single ptr
    Args:
        ptr: C pointer 
        size: size of array

    Returns:

    """
    IF SSE4_2:
        cdef double **a_ = <double **> memalign(16, size*sizeof(double*))
    ELSE:
        cdef double **a_ = <double **> malloc(size*sizeof(double*))

    if a_ is NULL:
        raise MemoryError("Failed to allocate memory!")

    cdef int i
    for i in range(size):
        a_[i] = ptr
        # printf("%f, %d\n", dereference(a_[i]), a_[i])
        # print(i, ref, dereference(a_[i]), hex(<unsigned long>&a_[i]))
        postincrement(ptr)

    return a_


cdef void get_array_from_ptr(double* ptr, double* a_, int size):

    cdef int i

    for i in range(size):
        a_[0] = dereference(ptr)
        postincrement(ptr)
        postincrement(a_)



from .chain_collection import Chain, ChainCollection
import numpy as np
from .chain import calculate_charge
from abpytools.utils import DataLoader
from .helper_functions import germline_identity_pd, to_numbering_table


class Fab:

    def __init__(self, heavy_chain=None, light_chain=None, load=True, name=None):
        """

        Args:
            heavy_chain:
            light_chain:
            load:
            name:
        """
        # As a convention the order will always be light chain (index 0) and then heavy chain (index 1)

        # check if it's an Chain class
        if heavy_chain is None and light_chain is None:
            raise ValueError('heavy_chain and light_chain must be provided')

        if isinstance(heavy_chain, Chain):
            self._heavy_chain = heavy_chain

        else:
            raise ValueError('heavy_chain must be a Chain object, but got {} instead'.format(type(heavy_chain)))

        if isinstance(light_chain, Chain):
            self._light_chain = light_chain

        else:
            raise ValueError('light_chain must be a Chain object, but got {} instead'.format(type(light_chain)))

        if self._heavy_chain.chain != 'heavy':
            raise ValueError("heavy_chain is not a heavy chain, it is {}".format(self._heavy_chain.chain))
        if self._light_chain.chain != 'light':
            raise ValueError("light_chain is not a light chain, it is {}".format(self._light_chain.chain))

        self._pair_sequence = self[0].sequence + self[1].sequence

        if isinstance(name, str):
            self._name = name

        elif name is None:
            self._name = 'ID1'

        # keep the name of the heavy and light chains internally to keep everything in the right order
        self._internal_heavy_name = self[1].name
        self._internal_light_name = self[0].name

    def load(self):
        self._heavy_chain.load()
        self._light_chain.load()

    def molecular_weight(self, monoisotopic=False):

        return self[0].ab_molecular_weight(monoisotopic=monoisotopic) +\
               self[1].ab_molecular_weight(monoisotopic=monoisotopic)

    def extinction_coefficient(self, reduced=False, normalise=False, **kwargs):

        light_ec = self[0].ab_ec(reduced=reduced, **kwargs)
        heavy_ec = self[1].ab_ec(reduced=reduced, **kwargs)

        if normalise:
            return (heavy_ec + light_ec) / (self.molecular_weight(**kwargs))
        else:
            return heavy_ec + light_ec

    def hydrophobicity_matrix(self, **kwargs):

        return np.concatenate((self[0].ab_hydrophobicity_matrix(**kwargs),
                               self[1].ab_hydrophobicity_matrix(**kwargs)))

    def charge(self, **kwargs):

        return np.concatenate((self[0].ab_charge(**kwargs), self[1].ab_charge(**kwargs)))

    def total_charge(self, ph=7.4, pka_database='Wikipedia'):

        available_pi_databases = ["EMBOSS", "DTASetect", "Solomon", "Sillero", "Rodwell", "Wikipedia", "Lehninger",
                                  "Grimsley"]
        assert pka_database in available_pi_databases, \
            "Selected pI database {} not available. Available databases: {}".format(pka_database,
                                                                                    ', '.join(available_pi_databases))

        data_loader = DataLoader(data_type='AminoAcidProperties', data=['pI', pka_database])
        pka_data = data_loader.get_data()

        return calculate_charge(sequence=self.sequence, ph=ph, pka_values=pka_data)

    # def load_igblast_query(self, file_path, chain):
    #
    #     if chain.lower() == 'light':
    #         self._light_chains.load_igblast_query(file_path=file_path)
    #     elif chain.lower() == 'heavy':
    #         self._heavy_chains.load_igblast_query(file_path=file_path)
    #     else:
    #         raise ValueError('Specify if the data being loaded is for the heavy or light chain')

    def numbering_table(self, as_array=False, region='all', chain='both'):

        return to_numbering_table(as_array=as_array, region=region, chain=chain,
                                  heavy_chains_numbering_table=self._heavy_chain.ab_numbering_table,
                                  light_chains_numbering_table=self._light_chain.ab_numbering_table,
                                  names=[self.name])

    @property
    def name(self):
        return self._name

    @property
    def sequence(self):
        return self._pair_sequence

    @property
    def aligned_sequence(self):
        return self[0].aligned_sequence + self[1].aligned_sequence

    @property
    def germline_identity(self):
        return self._germline_identity()

    # @property
    # def germline(self):
    #
    #     heavy_germline = self._heavy_chains.germline
    #     light_germline = self._light_chains.germline
    #
    #     return {name_i: {"Heavy": heavy_germline[heavy_i],
    #                      "Light": light_germline[light_i]} for name_i, heavy_i,
    #                                                            light_i in zip(self._names, self._internal_heavy_name,
    #                                                                           self._internal_light_name)}

    def _string_summary_basic(self):
        return "abpytools.Fab Name: {} Sequence length: {}".format(self.name, len(self.sequence))

    def __len__(self):
        return len(self.sequence)

    def __repr__(self):
        return "<%s at 0x%02x>" % (self._string_summary_basic(), id(self))

    def __getitem__(self, index):
        if index == 0:
            # get light chain
            return self._light_chain
        elif index == 1:
            return self._heavy_chain
        else:
            raise ValueError("Can only slice object with either 0 (light chain) or 1 (heavy chain)")

    def _germline_identity(self):

        # empty dictionaries return false
        if bool(self[1].germline_identity) is False:
            # this means there is no information about the germline,
            # by default it will run a web query
            # this is a very lazy fix to to do a web query using a Chain object...
            ChainCollection(antibody_objects=[self[1]]).igblast_server_query()
        if bool(self[0].germline_identity) is False:
            ChainCollection(antibody_objects=[self[0]]).igblast_server_query()

        return germline_identity_pd({self._internal_heavy_name: self[0].germline_identity},
                                    {self._internal_light_name: self[1].germline_identity},
                                    [self._internal_heavy_name],
                                    [self._internal_light_name],
                                    [self._name])



from .core import *
from .analysis.cdr_length import CDRLength
from .analysis.amino_acid_freq import AminoAcidFreq
from .analysis.sequence_alignment import SequenceAlignment
from .__about__ import __version__

VERSION = __version__



from . import ChainProto
from ..flags import *


def get_protobuf_numbering_scheme(numbering_scheme):
    """
    Returns ChainProto field value of a given numbering scheme
    Args:
        numbering_scheme:

    Returns:

    """
    if numbering_scheme == NUMBERING_FLAGS.KABAT:
        proto_numbering_scheme = ChainProto.KABAT
    elif numbering_scheme == NUMBERING_FLAGS.CHOTHIA:
        proto_numbering_scheme = ChainProto.CHOTHIA
    elif numbering_scheme == NUMBERING_FLAGS.CHOTHIA_EXT or numbering_scheme == NUMBERING_FLAGS.MARTIN:
        proto_numbering_scheme = ChainProto.CHOTHIA_EXT
    else:
        raise ValueError(f"{numbering_scheme} numbering scheme not supported by protobuf "
                         f"serialisation yet!")

    return proto_numbering_scheme


def get_numbering_scheme_from_protobuf(proto_numbering_scheme):
    """
    Returns ChainProto field value of a given numbering scheme
    Args:
        numbering_scheme:

    Returns:

    """
    if proto_numbering_scheme == ChainProto.KABAT:
        numbering_scheme = NUMBERING_FLAGS.KABAT
    elif proto_numbering_scheme == ChainProto.CHOTHIA:
        numbering_scheme = NUMBERING_FLAGS.CHOTHIA
    elif proto_numbering_scheme == ChainProto.CHOTHIA_EXT:
        numbering_scheme = NUMBERING_FLAGS.CHOTHIA_EXT
    else:
        raise ValueError(f"ChainProto numbering scheme {proto_numbering_scheme} is not compatible with abpytools!")

    return numbering_scheme



from setuptools import setup, Extension
import os
import sys
import shutil
from distutils.command.build_py import build_py as _build_py
from distutils.command.clean import clean as _clean
import re
import configparser

# Remove the "-Wstrict-prototypes" compiler option, which isn't valid for C++.
import distutils.sysconfig
import subprocess

cfg_vars = distutils.sysconfig.get_config_vars()
for key, value in cfg_vars.items():
    if type(value) == str:
        cfg_vars[key] = value.replace("-Wstrict-prototypes", "")

config = configparser.ConfigParser()

####################################################################
#                           VERSION
####################################################################
about = {}
with open('abpytools/__about__.py', 'r') as f:
    exec(f.read(), about)

####################################################################
#                            CYTHON
####################################################################

try:
    from Cython.Build import cythonize
    from Cython.Distutils import build_ext

    use_cython = True

    compile_time_env = {}


except ImportError:
    use_cython = False

    raise NotImplementedError("Currenly Cython is always required, but in future versions both "
                              "Python and Cython implementations will be available.")

if use_cython:

    cython_extensions = [Extension("abpytools.cython_extensions.convert_py_2_C",
                                   ["abpytools/cython_extensions/convert_py_2_C.pyx"],
                                   language='c++'),
                         Extension("abpytools.utils.math_utils",
                                   ["abpytools/utils/math_utils.pyx", "abpytools/utils/ops.cpp"],
                                   extra_compile_args=['-march=native', '-fopenmp', '-std=c++11'],
                                   extra_link_args=['-fopenmp'],
                                   language='c++'),
                         Extension("abpytools.analysis.distance_metrics_",
                                   ["abpytools/analysis/distance_metrics_.pyx"],
                                   language='c++')
                         ]

    # compile with SSE instructions?
    # TODO: check precise minimum SSE instruction set required
    try:
        call = subprocess.check_output("gcc -mavx2 -dM -E - < /dev/null | egrep 'SSE4_2'",
                                       stderr=subprocess.STDOUT, shell=True)
        compile_time_env['SSE4_2'] = 1
    except subprocess.CalledProcessError:
        compile_time_env['SSE4_2'] = 0

    cython_extensions_ = cythonize(cython_extensions,
                                   compile_time_env=compile_time_env,
                                   compiler_directives={'language_level': 3})


else:
    cython_extensions_ = None

####################################################################
#                           PROTOBUF
####################################################################
# adapted from https://github.com/protocolbuffers/protobuf/blob/master/python/setup.py
try:
    import google.protobuf as protobuf

    HAS_PROTOBUF = True
except:
    HAS_PROTOBUF = False

protoc = None


def get_protoc_version(path):
    """
    Get the protoc version
    Args:
        path: path to protoc

    Returns:
        protobuf version with the format "%d.%d.%d"
    """

    with subprocess.Popen(f"{path} --version", shell=True, stdout=subprocess.PIPE) as proc:
        # get stdout
        protobuf_version_b = proc.communicate()[0]
        protobuf_version_out = protobuf_version_b.decode("utf-8")
        protobuf_version_re = re.findall(r"\d.\d.\d", protobuf_version_out)
        if len(protobuf_version_re) == 1:
            protobuf_version = protobuf_version_re[0]
        else:
            protobuf_version = None
    return protobuf_version


if HAS_PROTOBUF:
    # Find the Protocol Compiler.
    if 'PROTOC' in os.environ and os.path.exists(os.environ['PROTOC']):
        candidate_protoc = os.environ['PROTOC']
        if get_protoc_version(candidate_protoc) == protobuf.__version__:
            protoc = candidate_protoc

    # look for protoc in anaconda first
    if protoc is None and os.path.exists('/'.join(sys.executable.split('/')[:-1] + ["protoc"])):
        candidate_protoc = '/'.join(sys.executable.split('/')[:-1] + ["protoc"])
        if get_protoc_version(candidate_protoc) == protobuf.__version__:
            protoc = candidate_protoc
    # look for protoc in /usr/local/bin
    if protoc is None and os.path.exists("/usr/local/bin/protoc"):
        candidate_protoc = "/usr/local/bin/protoc"
        if get_protoc_version(candidate_protoc) == protobuf.__version__:
            protoc = candidate_protoc
    if protoc is None:
        candidate_protoc = shutil.which("protoc")
        if get_protoc_version(candidate_protoc) == protobuf.__version__:
            protoc = candidate_protoc
    if protoc is None:
        print("-- Failed to find a matching proto compiler and protobuf python library!")
        print("-- Disabling protobuf support.")
        print(f"-- Protobuf python library version: {protobuf.__version__}")
        HAS_PROTOBUF = False
    else:
        print(f"-- Protoc  : {protoc} (Found {get_protoc_version(protoc)})\n"
              f"-- Protobuf: {protobuf.__file__} (Found {protobuf.__version__})")
        print("-- Found a matching proto compiler  and protobuf python library")
        HAS_PROTOBUF = True

if HAS_PROTOBUF:
    def generate_proto(source, require=True):
        """Invokes the Protocol Compiler to generate a _pb2.py from the given
        .proto file.  Does nothing if the output already exists and is newer than
        the input."""

        if not require and not os.path.exists(source):
            return

        output = source.replace(".proto", "_pb2.py")

        if (not os.path.exists(output) or
                (os.path.exists(source) and
                 os.path.getmtime(source) > os.path.getmtime(output))):
            print("Generating %s..." % output)

            if not os.path.exists(source):
                sys.stderr.write("Can't find required file: %s\n" % source)
                sys.exit(-1)

            if protoc is None:
                sys.stderr.write(
                    "protoc is not installed nor found in ../src.  Please compile it "
                    "or install the binary package.\n")
                sys.exit(-1)

            protoc_command = [protoc, "-I.", "--python_out=.", source]
            if subprocess.call(protoc_command) != 0:
                sys.exit(-1)


    class build_py(_build_py):
        def run(self):
            # Generate necessary .proto file if it doesn't exist.
            generate_proto("abpytools/core/formats/chain.proto")
            generate_proto("abpytools/core/formats/fab.proto")

            _build_py.run(self)

else:
    build_py = _build_py

if HAS_PROTOBUF:
    config['PROTOBUF'] = {'PROTOBUF': 1,
                          'PROTOC_VERSION': get_protoc_version(protoc),
                          'PROTOBUF_LIB_VERSION': protobuf.__version__}
else:
    config['PROTOBUF'] = {'PROTOBUF': 0,
                          'PROTOC_VERSION': 0,
                          'PROTOBUF_LIB_VERSION': 0}

with open("abpytools/config.ini", 'w') as configfile:
    config.write(configfile)


####################################################################
#                           CLEANUP
####################################################################
# adapted from https://stackoverflow.com/questions/27843481/python-project-using-protocol-buffers-deployment-issues
class clean(_clean):
    def run(self):
        # Delete generated files in the code tree.
        for (dirpath, dirnames, filenames) in os.walk("."):
            for filename in filenames:
                filepath = os.path.join(dirpath, filename)
                if filepath.endswith("_pb2.py") or filepath.endswith("so") or filepath.endswith("pyc"):
                    os.remove(filepath)
                if filepath.endswith(".pyx"):
                    # print(filepath.replace('pyx', 'cpp'))
                    try:
                        os.remove(filepath.replace('pyx', 'cpp'))
                    except:
                        # cython generated file no longer exists
                        pass
        # _clean is an old-style class, so super() doesn't work.
        _clean.run(self)


####################################################################
#                           SETUP
####################################################################
setup(
    name='AbPyTools',
    classifiers=['Development Status :: 2 - Pre-Alpha',
                 'Intended Audience :: Science/Research',
                 'License :: OSI Approved :: MIT License',
                 'Natural Language :: English',
                 'Programming Language :: Python :: 3 :: Only',
                 'Topic :: Scientific/Engineering :: Bio-Informatics'],
    keywords='antibody-analysis bioinformatics data-processing',
    version=about['__version__'],
    package_dir={'abpytools': 'abpytools'},
    packages=['abpytools',
              'abpytools.utils',
              'abpytools.core',
              'abpytools.core.flags',
              'abpytools.core.formats',
              'abpytools.analysis',
              'abpytools.features',
              'abpytools.cython_extensions'],
    package_data={'abpytools': ['data/*.json',
                                'config.ini']},
    url='https://github.com/gf712/AbPyTools',
    license=about['__license__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    description='Python package for antibody analysis',
    install_requires=['numpy',
                      'joblib',
                      'pandas',
                      'tqdm>=4.15',
                      'seaborn',
                      'matplotlib',
                      'beautifulsoup4',
                      'lxml',
                      'scipy',
                      "parameterized"],
    test_suite="tests",
    ext_modules=cython_extensions_,
    cmdclass={'build_ext': build_ext,
              'build_py': build_py,
              'clean': clean}
)



from libc.math cimport pow as pow_C
from abpytools.cython_extensions.convert_py_2_C cimport (get_C_double_array_pointers, release_C_pointer, memalign)
import itertools
from cython.operator cimport dereference, postincrement
from libc.stdlib cimport malloc, free

cdef extern from "ops.h":
    double norm_op(double *A, int p, int size)
    void exp_op(double *A, double *B, int size)
    double l2_norm_op(double *A, int size)
    double l1_norm_op(const double *A, int size)
    # elementwise ops
    void subtract_op(double *A, double *B, double *C, int size)
    void multiply_op(double *A, double *B, double *C, int size)
    void add_op(double *A, double *B, double *C, int size)


cdef class NumericalBaseStructure:
    pass


cdef class Matrix2D_backend:
    """
    Cython class to manipulate double precision 2D matrix
    """

    def __init__(self, values_, ptrs=None, shape=None):
        # if ptr and shape are defined, values will be ignored and the data will be read from ptrs
        # check all rows have the same size
        self.values_ = values_
        self.n_rows = len(self.values_)
        self.n_cols = len(self.values_[0])
        self.size_ = self.n_cols * self.n_rows

        cdef int i
        for i in range(1, self.n_rows):
           if len(self.values_[i]) != self.n_cols:
               raise ValueError("All rows must have the same number of elements")

        # copy data to a contiguous C array
        self.data_C = get_C_double_array_pointers(list(itertools.chain(*self.values_)), self.size_)

    def _check_index(self, int row, int column):
        if row * self.n_cols + column < self.size_:
            raise IndexError("Requested element is outside array bounds")

    cdef double* _get_value_pointer(self, int row, int col):
        cdef double* ptr = &self.data_C[self.n_cols * row + col]

    cdef double* _get_row_pointer(self, int row):
        """
        Get pointer of first element of this row
        Args:
            row:

        Returns:

        """
        # pointer to first element of C array row
        cdef double* ptr = &self.data_C[self.n_cols * row]
        # print(dereference(ptr), hex(<unsigned long>&ptr))
        return ptr

    cpdef list _get_row(self, int row):
        result = []
        cdef int i
        cdef double* ptr = self._get_row_pointer(row)
        for i in range(self.n_cols):
            result.append(dereference(ptr))
            postincrement(ptr)

        return result

    def __dealloc__(self):
        # releases C memory allocation
        # print("Releasing Matrix memory ({})".format(id(self)))
        release_C_pointer(self.data_C)
        # release_C_pp(self.data_C)

    cdef double _get_value(self, int row, int col):
        return dereference(self._get_row_pointer(row)+col)

    cdef void _set_value(self, int row, int col, double value):
        cdef double* ptr =  self._get_row_pointer(row) + col
        ptr[0] = value


    def __getitem__(self, idx):
        if isinstance(idx, tuple) and len(idx) == 2:
            # return a single value (only supports rows and columns)
            # NOTE that there are no bound checks, so this can lead to odd behaviour,
            # e.g. asking for a[0, a.shape[1]+1] will return the first element of the second row
            # however this avoids wasting resources creating a new Vector object to get a single element
            row, column = idx
            return self._get_value(row, column)
        else:
            # create vector using data_ pointers
            return Vector.create(self._get_row_pointer(idx), self.n_cols)

    def __setitem__(self, idx, value):
        if isinstance(idx, tuple) and len(idx) == 2:
            row, column = idx
            self._set_value(row, column, value)
        else:
            raise NotImplementedError("Cannot assign vectors to rows in Matrix2D yet!")


    cpdef list _get_values(self):
        """
        Returns all values of C array
        Returns:

        """
        cdef list result = []
        cdef int i
        cdef int j
        cdef double* ptr
        cdef list row_result

        for i in range(self.n_rows):
            row_result = []
            ptr = self._get_row_pointer(i)  # this is more of a sanity check than anything else
            for j in range(self.n_cols):
                row_result.append(dereference(ptr))  # same as *ptr in C
                postincrement(ptr) # same as ptr++ in C
            result.append(row_result)

        return result



class Matrix2D(Matrix2D_backend):

    def __init__(self, values):
        """
        Lightweight numpy-like class for fast numerical calculations with Cython
        Args:
            values:
        """
        super().__init__(values)

    @property
    def values(self):
        """
        Returns data as a list of lists
        """
        return self._get_values()

    @property
    def shape(self):
        return self.n_rows, self.n_cols

    def __str__(self):

        final_string = "[{}]"
        intermediate_string = ""
        for i in range(self.n_rows):
            intermediate_string += "[{}]".format(', '.join([str(x) for x in self._get_row(i)]))

        return final_string.format(intermediate_string)

    def __repr__(self):

        repr_string = "array"
        first=True
        for x in self.__str__().split('\n'):
            if first:
                repr_string += "{}\n".format(x)
                first=False
            else:
                repr_string+="     {}\n".format(x)
        return repr_string


cdef class Vector:

    """
    Lightweight Cython vector class implementation that stores data as C arrays and uses these to perform calculations
    """

    def __init__(self, values_=None):
        if values_ is not None:
            self.size_ = len(values_)
            self.data_C = get_C_double_array_pointers(values_, self.size_)
            self.derived_=0


    @staticmethod
    cdef Vector create(double* ptr, int size):
        """
        Create a Vector from a pointer.
        
        Args:
            ptr: 
            size: 

        Returns:

        """
        cdef Vector vec = Vector()
        vec.size_ = size
        vec.data_C = ptr
        vec.derived_=1
        return vec

    @staticmethod
    cdef Vector allocate(int size):
        """
        
        Args:  
            size: 

        Returns:

        """
        IF SSE4_2:
            cdef double *value_pointers_ = <double *> memalign(16, size*sizeof(double))
        ELSE:
            cdef double *value_pointers_ = <double *> malloc(size*sizeof(double))

        if value_pointers_ is NULL:
            raise MemoryError("Failed to allocated memory")

        cdef Vector vec = Vector()
        vec.size_=size
        vec.data_C = value_pointers_
        vec.derived_=0

        return vec

    @staticmethod
    cdef Vector allocate_with_value(int size, double value):
        cdef Vector vec = Vector.allocate(size)
        cdef double *value_ptr = vec.data_C
        for i in range(size):
            value_ptr[0] = value
            postincrement(value_ptr)
        vec.derived_=0
        return vec


    cpdef double dot_product(self, Vector other):

        """
        Vector dot product.
        Args:
            other (Vector): a Vector object to calculate the dot product with

        Returns: the dot product between two vectors

        """

        if self.size_ != other.size:
            raise ValueError("Vector size mismatch")

        return internal_vector_dot_product_(self.data_C, other.data_C, self.size_)

    cdef double _get_element(self, int idx):
        return self.data_C[idx]

    cdef void _set_array_value(self, int idx, double value):
        cdef double* ptr = &self.data_C[idx]
        ptr[0] = value

    cpdef Vector exp(self):
        cdef Vector vec = Vector.allocate(self.size_)
        exp_op(self.data_C, vec.data_C, self.size_)
        return vec

    cpdef double norm(self, int p=2):
        if p==2:
            return l2_norm_op(self.data_C, self.size_)
        elif p == 1:
            return l1_norm_op(self.data_C, self.size_)
        else:
            return internal_vector_norm_(self.data_C, p, self.size_)

    @property
    def values(self):
        """
        Returns:
        """
        return [self._get_element(i) for i in range(self.size_)]

    def __dealloc__(self):
        if not self.derived_:
            if self.data_C != NULL:
                free(self.data_C)

    @property
    def size(self):
        return self.size_

    def __getitem__(self, int idx):
        return self._get_element(idx)

    def __setitem__(self, idx, value):
        self._set_array_value(idx, value)

    cpdef Vector subtract(self, Vector other):
        """
        Elementwise subtraction between two vectors.
        
        Args:
            other: 

        Returns:

        """

        # create empty vector
        cdef Vector vec = Vector.allocate(self.size_)

        subtract_op(self.data_C, other.data_C, vec.data_C, self.size_)

        return vec


    cpdef Vector add(self, Vector other):
        """
        Elementwise addition between two vectors.
        
        Args:
            other: 
    
        Returns:
    
        """

        # create empty vector
        cdef Vector vec = Vector.allocate(self.size_)

        add_op(self.data_C, other.data_C, vec.data_C, self.size_)

        return vec


    cpdef Vector multiply(self, Vector other):
        """
        Elementwise multiplication between two vectors.
        
        Args:
            other: 

        Returns:

        """
        # create empty vector
        cdef Vector vec = Vector.allocate(self.size_)

        multiply_op(self.data_C, other.data_C, vec.data_C, self.size_)

        return vec

    def __add__(self, other):
        return self.add(other)

    def __sub__(self, other):
        return self.subtract(other)

    def __mul__(self, other):
        return self.multiply(other)

    def __iter__(self):
        for x in range(self.size_):
            yield self[x]


cdef double internal_vector_dot_product_(double *u, double *v, int size):
    """
    "Pure" C definition of dot product for Cython compiler (to be only used in the backend).
    
    Args:
        u: 
        v: 
        size: 

    Returns:

    """

    cdef int i
    cdef double result = 0.0

    for i in range(size):
        result += u[i] * v[i]

    return result


cdef double internal_vector_norm_(double *u, int norm, int size):
    """
    "Pure" C definition of vector norm for Cython compiler (to be only used in the backend).
    
    Args:
        u: 
        norm: 
        size: 

    Returns:

    """

    cdef int i
    cdef double result = 0.0
    cdef double inverse_norm = 1.0 / norm

    for i in range(size):
        result += pow_C(u[i], norm)

    result = pow_C(result, inverse_norm)

    return result


cpdef double dot_product_(list u, list v):
    """
    Python API to use Cython dot_product implementation.
    
    Args:
        u list: Python list representing vector u  
        v list: Python list representing vector v

    Returns: Dot product of u and v

    """

    cdef int size
    if len(u) == len(v):
        size = len(u)
    else:
        raise ValueError("Vector size mismatch")

    cdef double *u_ = get_C_double_array_pointers(u, size)
    cdef double *v_ = get_C_double_array_pointers(v, size)

    cdef double result = internal_vector_dot_product_(u_, v_, size)

    release_C_pointer(u_)
    release_C_pointer(v_)

    return result



import json
from abpytools.home import Home


class DataLoader:

    def __init__(self, data=[], data_type='', amino_acid_property=[]):

        """

        :param data: list with the information required to access data in json file
        :param data_type: str must be one of the following: 'CDR_positions', 'NumberingSchemes' or 'AminoAcidProperties'
        :param amino_acid_property: temporary parameter
        :param misc: temporary parameter
        """
        self.amino_acid_property = amino_acid_property
        self.directory_name = Home().homedir
        self.data = data
        self.data_types = ['CDR_positions', 'NumberingSchemes', 'AminoAcidProperties', 'Framework_positions']

        if data_type not in self.data_types:
            raise ValueError("{} is not a valid data type. Available data types: {}".format(
                data_type, ', '.join(self.data_types)))
        else:
            self.data_type = data_type

        # checks values when object is instantiated
        if self.data_type == 'CDR_positions' or self.data_type == 'NumberingSchemes' \
                or self.data_type == 'Framework_positions':
            if len(self.data) != 2:
                raise ValueError("Expected 2, instead of {} values.".format(len(self.data)))
            if self.data[0] not in ['chothia', 'kabat', 'chothia_ext']:
                raise ValueError("Got {}, but only {} is available".format(self.data[0], 'chothia'))
            if self.data[1] not in ["light", "heavy"]:
                raise ValueError("Got {}, but only light and heavy are available".format(self.data[1]))
        else:
            if len(self.data) != 2:
                raise ValueError("Expected 2, instead of {} values.".format(len(self.data)))
            if self.data[0] not in ["hydrophobicity", "pI", "MolecularWeight", "ExtinctionCoefficient"]:
                raise ValueError("Got {}, but only {} are available".format(self.data[0],
                                                                            """hydrophobicity, pI, ExtinctionCoefficient
                                                                             and MolecularWeight"""))

            if self.data[1] not in ["kdHydrophobicity", "wwHydrophobicity", "hhHydrophobicity", "mfHydrophobicity",
                                    "ewHydrophobicity", "EMBOSS", "DTASetect", "Solomon", "Sillero", "Rodwell",
                                    "Wikipedia", "Lehninger", "Grimsley", "average", "monoisotopic", "Standard",
                                    "Standard_reduced"]:
                raise ValueError("Got {}, but only {} are available".format(self.data[1],
                                                                            """
                                                                            kdHydrophobicity ,wwHydrophobicity,
                                                                            hhHydrophobicity, mfHydrophobicity,
                                                                            ewHydrophobicity, EMBOSS, DTASetect,
                                                                            Solomon, Sillero, Rodwell,
                                                                            Wikipedia, Lehninger, Grimsley,
                                                                            average, monoisotopic, Standard and
                                                                            Standard_reduced
                                                                            """
                                                                            ))

    def get_data(self):

        if self.data_type == 'CDR_positions':
            with open('{}/data/CDR_positions.json'.format(self.directory_name), 'r') as f:
                # need to access numbering scheme and chain type
                data = json.load(f)[self.data[0]][self.data[1]]
        elif self.data_type == 'Framework_positions':
            with open('{}/data/Framework_positions.json'.format(self.directory_name), 'r') as f:
                # need to access numbering scheme and chain type
                data = json.load(f)[self.data[0]][self.data[1]]
        elif self.data_type == 'NumberingSchemes':
            with open('{}/data/NumberingSchemes.json'.format(self.directory_name), 'r') as f:
                # need to access numbering scheme and chain type
                data = json.load(f)[self.data[0]][self.data[1]]
        else:
            with open('{}/data/AminoAcidProperties.json'.format(self.directory_name), 'r') as f:
                data = json.load(f)[self.data[0]][self.data[1]]

        return data



from urllib import request, error


class Download:
    def __init__(self, url='', verbose=False, timeout=5):
        self.url = url
        self.verbose = verbose
        self.html = ''
        self.error = False
        self.timeout = timeout

    def download(self, user_agent='wswp', num_retries=2):
        # self.html, self.error = download(self.url, self.verbose, user_agent=user_agent, num_retries=num_retries,
        #                                  timeout=self.timeout)
        try:
            self.html = download(self.url, self.verbose, user_agent=user_agent, num_retries=num_retries,
                                 timeout=self.timeout)
        except IOError:
            raise ValueError("Could not download requested page.")


def download(url, verbose, user_agent='wswp', num_retries=2, decoding_format='utf-8', timeout=5):
    """
    Function to download contents from a given url

    Input:
            url: str
            string with the url to download from

            user_agent: str
            Default 'wswp'

            num_retries: int
            Number of times to retry downloading
            if there is an error

            verbose: bool
            Print out url and errors

            decoding: "utf-8"

    Output:
            returns: str
            string with contents of given url
    """

    # html_error = False
    if verbose:
        print('Downloading:', url)
    headers = {'User-agent': user_agent}
    request_obj = request.Request(url, headers=headers)
    try:
        with request.urlopen(request_obj, timeout=timeout) as response:
            html = response.read()
    except error.URLError as e:
        if verbose:
            print('Download error:', e.reason)
        # html = None
        # if num_retries > 0:
        #     if hasattr(e, 'code') and 500 <= e.code < 600:
        #         # retry 5XX HTTP errors
        #         return download(url, user_agent, num_retries - 1)[0]
        #     # elif hasattr(e, 'code') and e.code == 404:
        #     else:
        #         html_error = True
        raise IOError(e.reason)

    return html.decode(decoding_format)



from abpytools.core.flags import numbering, pi, backend, options, hydrophobicity, chain, formats

# create alias to flag modules
NUMBERING_FLAGS = numbering
CHAIN_FLAGS = chain
OPTION_FLAGS = options
PI_FLAGS = pi
BACKEND_FLAGS = backend
HYDROPHOBICITY_FLAGS = hydrophobicity
FORMAT_FLAGS = formats

# from Python 3.7 can do this
# import abpytools.core.flags.numbering as NUMBERING_FLAGS
# import abpytools.core.flags.chain as CHAIN_FLAGS
# import abpytools.core.flags.options as OPTION_FLAGS
# import abpytools.core.flags.pi as PI_FLAGS
# import abpytools.core.flags.hydrophobicity as HYDROPHOBICITY_FLAGS
# import abpytools.core.flags.backend as BACKEND_FLAGS

__all__ = ["NUMBERING_FLAGS",
           "CHAIN_FLAGS",
           "OPTION_FLAGS",
           "PI_FLAGS",
           "HYDROPHOBICITY_FLAGS",
           "BACKEND_FLAGS",
           "FORMAT_FLAGS"]



from matplotlib import pyplot as plt
import seaborn as sns
import os
from abpytools.utils import PythonConfig
from abpytools.features.regions import ChainDomains
from matplotlib.ticker import MaxNLocator
from .analysis_helper_functions import switch_interactive_mode


class CDRLength(ChainDomains):

    def __init__(self, path=None, antibody_objects=None, verbose=True, show_progressbar=True, n_threads=10):

        super().__init__(path=path, antibody_objects=antibody_objects, verbose=verbose,
                         show_progressbar=show_progressbar, n_threads=n_threads)

    def plot_cdr(self, only_cdr3=True, save=False, plot_path='./', plot_name='CDR_length',
                 plot_title=None, hist=True, ax=None, **kwargs):

        switch_interactive_mode(save=save)

        if ax is None:
            if only_cdr3:
                f, ax = plt.subplots(nrows=1, ncols=1)
            else:
                f, ax = plt.subplots(nrows=1, ncols=3, figsize=(15, 5))
                ax = ax.ravel()

        if only_cdr3:
            if plot_title is None:
                ax.set_title('CDR3 Length', size=18)
            else:
                ax.set_title(plot_title, size=18)
            sns.distplot(self.cdr_lengths()[:, 2], hist=hist, ax=ax, **kwargs)
            ax.set_ylabel('Density', size=14)
            ax.set_xlabel('CDR Length', size=14)
            ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        else:
            if plot_title is None:
                plt.suptitle('CDR Length', size=20)
            else:
                plt.suptitle(plot_title, size=20)
            for i, cdr in enumerate(['CDR 1', 'CDR 2', 'CDR 3']):

                ax[i].set_title(cdr, size=16)
                sns.distplot(self.cdr_lengths()[:, i], hist=hist, ax=ax[i])

                if i == 0:
                    ax[i].set_ylabel('Density', size=16)

                if i == 1:
                    ax[i].set_xlabel('CDR Length', size=16)

                ax[i].xaxis.set_major_locator(MaxNLocator(integer=True))

            plt.tight_layout()
            plt.subplots_adjust(top=0.85)

        ipython_config = PythonConfig()
        if ipython_config.ipython_info == 'notebook' and save is False:
            plt.plot()
        else:
            plt.savefig(os.path.join(plot_path, plot_name), format='png')
            plt.close()



import configparser
from abpytools.home import Home

config = configparser.ConfigParser()
config.read(f"{Home().homedir}/config.ini")

HAS_PROTO = True if config["PROTOBUF"]['protobuf'] == '1' else False
